require conv_main;
require conv_check;

$input_file_handle;

$DEFAULT_CONVERGENCE_LIMIT = 0.00001;
$DEFAULT_ITERATION_LIMIT = 100;
$DEFAULT_PRINT_INTERVAL = 1;
$DEFAULT_UNDERRELAXATION_COEFF = 0.9;

sub open_input_file
{
# $_[0]: contains the filename of the input file to read

   if (open (INPUT_FILE, $_[0]) != 1)
   {
      die "ERROR - could not open file \"$_[0]\" for reading.\n"; 
   }

   $input_file_handle = *INPUT_FILE;
}


sub close_input_file
{
   close($input_file_handle);
}


sub parse_initial_information_LQN
{
# First thing is to rewind to the beginning of the file

   seek($input_file_handle, 0, 0);
   $cur_line = 0;

# A robust algorithm below to ensure we get everything between the quotes, so the user
# can put any number of -1 in between the quotes and not confuse the parser.

# regexp \s*\-1\$ fails when you have a lqn parameter of 1e-1, SPEX fails the same way...

   $preamble_end = 0;

   while ($preamble_end == 0)
   {
      $last_line = get_next_line_until(0, "\\s*\\-1\$", \$model_preamble);
      $model_preamble = $model_preamble.$last_line;

      if (($last_line =~ /^-1/) || ($last_line =~ /\s+-1/))
      {
         if ($model_preamble =~ /\".*\"/s)
         {
            $preamble_end = 1;
	 }
      }
   }

   @preamble_split = split("\"", $model_preamble);

# So now [1] will contain all the stuff within the quotes, [2] will contain the 4 LQN parameters

   $model_comments = @preamble_split[1];

   @preamble_split[2] =~ s/\#.*\n//g;
   @preamble_split[2] =~ s/^\s+//g;
   @lqn_solver_params = split(/\s+/, @preamble_split[2]);
   $num_lqn_solver_params = @lqn_solver_params;

   if ($num_lqn_solver_params != 5)  # Note that includes the -1
   {
      die "ERROR - There is something wrong in the pre-amble part of the model file.  Specifically the problem is after the model comment, and before the -1.\n";
   }

   $convergence_limit = @lqn_solver_params[0];
   $iteration_limit = @lqn_solver_params[1];
   $print_interval = @lqn_solver_params[2];
   $underrelaxation_coeff = @lqn_solver_params[3];
}


sub parse_initial_information_SPEX
{
   my $line, num_temp_var;
   my @temp_var = ();

# set some default values...

   $convergence_limit = $DEFAULT_CONVERGENCE_LIMIT;
   $iteration_limit = $DEFAULT_ITERATION_LIMIT;
   $print_interval = $DEFAULT_PRINT_INTERVAL;
   $underrelaxation_coeff = $DEFAULT_UNDERRELAXATION_COEFF;

# First thing is to rewind to the beginning of the file

   seek($input_file_handle, 0, 0);
   $stop_parsing = 0;
   $cur_line = 0;

   my $prev_file_pos = tell($input_file_handle);
   my $prev_cur_line = 0;

   while ($stop_parsing == 0)
   {
      chop($line = <$input_file_handle>);

# replace all tabs characters with spaces

      $line =~ s/\t/ /g;

# strip out the CTRL-M's if they exist.  DOS/Windows like to put a CTRL-M at the end of each line

      $line =~ s/\015//;
       
# check to see if the beginning of the line starts with a space, if so strip off all spaces at the beginning of $line.
      if (index($line, " ") == 0)
      {
         $line =~ s/ +//;
      }

      $cur_line++;

      if ($line =~ /^\s*\$convergence_limit\s*=\s*([0-9\.]+)/)
      {
         $convergence_limit = $1;
      }
      elsif ($line =~ /^\s*\$solver\s*=\s*(.*)$/)
      {
         @temp_var = split(" ", $1);
         $num_temp_var = @temp_var;
         $solver = @temp_var[0];
         $solver_parameters = join(" ", @temp_var[1..$num_temp_var]);
      }
      elsif ($line =~ /^\s*\$iteration_limit\s*=\s*([0-9\.]+)/)
      {
         $iteration_limit = $1;
      }
      elsif ($line =~ /^\s*\$print_interval\s*=\s*([0-9\.]+)/)
      {
         $print_interval = $1;
      }
      elsif ($line =~ /^\s*\$coefficient_of_variation\s*=\s*([0-9\.]+)/)
      {
         $underrelaxation_coeff = $1;
      }
      elsif (($line =~ /^\s*pragma\s+(.*)$/) || ($line =~ /^\s*\#pragma\s+(.*)$/))
      {
         $temp_line = $1;
         $temp_line =~ s/\s*#.*$//g;    # stip out any comments on this line
         $model_pragmas = $model_pragmas."$temp_line\n";
      }
      elsif ($line =~ /^\s*(\$.*)$/)
      {
         $SPEX_init_block = $SPEX_init_block.$1."\n";
      }
      else
      {
         if ($line =~ /^\s*P\s+[0-9]+$/)
         {
            $stop_parsing = 1;
            seek($input_file_handle, $prev_file_pos, 0);
            $cur_line = $prev_cur_line;
	 }
	 else
         {
             if (($line !~ /^\s*\#.*$/) && ($line =~ /.+/))
             {
	        die "ERROR - file appears to be of incorrect format.  Needs to be an lqn or xlqn file.\n";
	     }

             $model_comments = $model_comments.$line."\n";
	 }
      }

      $prev_file_pos = tell($input_file_handle);
      $prev_cur_line = $cur_line;
   }

   if ($debug_msg > 0)
   {
      print "*************\n$model_comments\n-----\n";
      print "conv_limit: $convergence_limit\n";
      print "it-lim: $iteration_limit\n";
      print "print_inv: $print_interval\n";
      print "under: $underrelaxation_coeff\n";
      print "pragmas: $model_pragmas\n";
   }
}
       

sub parse_processor_block
{
   my @SPEX_observations = ();
   @proc_tokens = &get_next_line(0, \@SPEX_observations);
   $num_proc_tokens = @proc_tokens;

   while (@proc_tokens[0] ne "-1")
   {
      if (@proc_tokens[0] ne "p")
      {
         die "ERROR at line $cur_line: Expecting a \"p\" at the beginning of the line.\n";
      }

      if ($num_proc_tokens < 3)
      {
	 die "ERROR at line $cur_line: There seems to be insufficient amount of information describing this particular processor.\n";
      }
      
      &check_valid_identifier_name(@proc_tokens[1]);

      if (defined $processor_list{@proc_tokens[1]})
      {
         die "ERROR at line $cur_line - the PROCESSOR name @proc_tokens[1] has already been used.\n";
      }

   # Set the default values in the process_list entry
   # structure for a processor entry:
   # pos 0:discipline, 1:multiplicity, 2:quantum, 3:replicas, 4:proc_rate
  
   # Set default values

      my $multiplicity = 1;
      my $quantum = 0;
      my $replicas = 0;
      my $proc_rate = 1;
 
      &init_ordering(0);

      $x=3;
   
      $temp_value = @proc_tokens[$x];

      if (($temp_value =~ /^[0-9]+/) || ($temp_value =~ /^\$/))
      {
         &check_ordering(1);
         &check_valid_real(@proc_tokens[$x]);
         $quantum = @proc_tokens[$x];
         $x++;
      }

      if (@proc_tokens[$x] eq "i")
      {
         &check_ordering(2);
	 $multiplicity = 0;
	 $x++;
      }

      if (@proc_tokens[$x] eq "m")
      {
         &check_ordering(3);
         &check_valid_integer(@proc_tokens[$x+1]);
	 $multiplicity = @proc_tokens[$x+1];
	 $x=$x+2;
      }

      if (@proc_tokens[$x] eq "r")
      {
         &check_ordering(4);
         &check_valid_integer(@proc_tokens[$x+1]);
	 $replicas = @proc_tokens[$x+1];
	 $x=$x+2;
      }

      if (@proc_tokens[$x] eq "R")
      {
         &check_ordering(5);
         &check_valid_real(@proc_tokens[$x+1]);
	 $proc_rate = @proc_tokens[$x+1];
	 $x=$x+2;
      }

      if ($x != $num_proc_tokens)
      {
         die "ERROR at line $cur_line: An invalid parameter is being used, or some parameters are out of order in the processor specification.\n";
      }

   # pos 0:discipline, 1:multiplicity, 2:quantum, 3:replicas, 4:proc_rate
      $processor_list{@proc_tokens[1]} = "@proc_tokens[2] $multiplicity $quantum $replicas $proc_rate";

      if (($op_mode eq "SPEX") && ($#SPEX_observations != -1))
      {
         &add_SPEX_observation_variables(\@SPEX_observations);
      # This $cur_processor is a reminisent of the spreadconv code, get rid of it!!!!!!!!!
         $cur_processor = @proc_tokens[1];
         check_SPEX_observations(\@SPEX_observations, "processor");

         $processor_list{@proc_tokens[1]} = $processor_list{@proc_tokens[1]}.";";

      # tack on the SPEX observation keys:

         foreach $cur_obs (@SPEX_observations)
         {
            $processor_list{@proc_tokens[1]} = $processor_list{@proc_tokens[1]}."%".$cur_obs." ";
         }
      }

      if ($debug_msg > 0)
      {
         print "Found processor @proc_tokens[1], internal representation: $processor_list{@proc_tokens[1]}\n";
      }

      @proc_tokens = &get_next_line(0, \@SPEX_observations);
      $num_proc_tokens = @proc_tokens;
   }
}

$cur_order = 0;

sub init_ordering
{
   $cur_order = $_[0];
}

sub check_ordering
{
# $_[0] this is the current order

   if ($cur_order >= $_[0])
   {
      die "ERROR at line $cur_line: Parameters on this line appear out of order or duplicated, please correct this.\n";
   }
   else
   {
     $cur_order = $_[0];
   }
}


sub parse_task_block
{
   my @SPEX_observations = ();
   @task_tokens = &get_next_line(0, \@SPEX_observations);
   $num_task_tokens = @task_tokens;

   while (@task_tokens[0] ne "-1")
   {

# setting default values below

      my $processor = "none";
      my $task_priority = 0;
      my $think_time = 0;
      my $multiplicity = 1;
      my $replication = 0;

      if (@task_tokens[0] ne "t")
      {
         die "ERROR at line $cur_line: Expecting a \"t\" at the beginning of the line.\n";
      }

      if (defined $task_list1{@task_tokens[1]})
      {
	 die "ERROR at line $cur_line - the TASK name @task_tokens[1] has already been used.\n";
      }

      if ($num_task_tokens < 5)
      {
	 die "ERROR at line $cur_line: There seems to be insufficient amount of information describing this particular task.\n";
      }
      
      &check_valid_identifier_name(@task_tokens[1]);

      if (defined $task_list1{@task_tokens[1]})
      {
         die "ERROR at line $cur_line - the TASK name @task_tokens[1] has already been used.\n";
      }

   # must traverse the entry list defined in the task

      $x = 3;
      $task_list1{@task_tokens[1]} = @task_tokens[2]." ";

      while (@task_tokens[$x] ne "-1")
      {
         &check_valid_identifier_name(@task_tokens[$x]);
         $task_list1{@task_tokens[1]} = $task_list1{@task_tokens[1]}.@task_tokens[$x]." ";
         $x++;
      }

      $x++;    # increment over the -1 for the end of a entry list
      $processor = @task_tokens[$x];

# if the processor is a SPEX variable, we can't do much checking to see if the task is assigned to a valid processor.

      if (($op_mode eq "SPEX") && !($processor =~ /^\$/))
      {
         if (!defined $processor_list{$processor})
         {
            die "ERROR at line $cur_line: Processor $processor has been defined in the task block, but not in the processor block.\n";
         }
      }

      if ($processor =~ /^\$/)
      {
	  print "WARNING - Task \"@task_tokens[1]\" will not be converted into the spreadsheet because it was assigned to a SPEX variable, not a concrete processor.\n";
      }

      &init_ordering(0);
      $x++;
   
      $temp_value = @task_tokens[$x];

      if (($temp_value =~ /^[0-9]+$/) || ($temp_value =~ /^\$/))
      {
         &check_ordering(1);
         &check_valid_integer(@task_tokens[$x]);
         $task_priority = @task_tokens[$x];
         $x++;
      }

      if (@task_tokens[$x] eq "z")
      {
         &check_ordering(2);
         &check_valid_real(@task_tokens[$x+1]);
	 $think_time = @task_tokens[$x+1];
	 $x=$x+2;
      }

      if (@task_tokens[$x] eq "i")
      {
         &check_ordering(3);
	 $multiplicity = 0;
	 $x++;
      }

      if (@task_tokens[$x] eq "m")
      {
         &check_ordering(4);
         &check_valid_integer(@task_tokens[$x+1]);
	 $multiplicity = @task_tokens[$x+1];
	 $x=$x+2;
      }

      if (@task_tokens[$x] eq "r")
      {
         &check_ordering(5);
         &check_valid_integer(@task_tokens[$x+1]);
	 $replication = @task_tokens[$x+1];
	 $x=$x+2;
      }

      if ($x != $num_task_tokens)
      {
         die "ERROR at line $cur_line: An invalid parameter is being used, or some parameters are out of order in the task specification.\n";
      }

   # structure for a task entry: (task_list2, note that discipline is stored in task_list1)
   # pos 0:processor, 1:priority, 2:think time, 3:multiplicity, 4:replicas

      $task_list2{@task_tokens[1]} = $processor." ".$task_priority." ".$think_time." ".$multiplicity." ".$replication;

      if (($op_mode eq "SPEX") && ($#SPEX_observations != -1))
      {
         &add_SPEX_observation_variables(\@SPEX_observations);

      # This $cur_task is a reminisent of the spreadconv code, get rid of it!!!!!!!!!
         $cur_task = @task_tokens[1];
         check_SPEX_observations(\@SPEX_observations, "task");

         $task_list2{@task_tokens[1]} = $task_list2{@task_tokens[1]}.";";

      # tack on the SPEX observation keys:

         foreach $cur_obs (@SPEX_observations)
         {
            $task_list2{@task_tokens[1]} = $task_list2{@task_tokens[1]}."%".$cur_obs." ";
         }
      }

      if ($debug_msg > 0)
      {
         print "Found task @task_tokens[1], internal representation: task_list1: $task_list1{@task_tokens[1]} task_list2: $task_list2{@task_tokens[1]}\n";
      }

      @task_tokens = &get_next_line(0, \@SPEX_observations);
      $num_task_tokens = @task_tokens;
  }
}


sub parse_entry_block
{
   my @SPEX_observations = ();
   @entry_tokens = &get_next_line(0, \@SPEX_observations);
   $num_entry_tokens = @entry_tokens;

#   &add_SPEX_observation_variables(\@SPEX_observations);

   while (@entry_tokens[0] ne "-1")
   {
      if (@entry_tokens[0] eq "a")
      {
         if ($num_entry_tokens != 3)
         {
            die "ERROR on line $cur_line - There seems to be too much or too little information in specify the entry open arrival rate.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_real(@entry_tokens[2]);

         $entry_list_arriv_rate{@entry_tokens[1]} = @entry_tokens[2];
      
         if ($debug_msg > 0)
         {
            print "Entry arriv_rate list <@entry_tokens[1]>  $entry_list_arriv_rate{@entry_tokens[1]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "c")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify coefficient of variation.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_phase_parameters(2, \@entry_tokens, \&check_valid_real);

         $entry_list_COV{@entry_tokens[1]} = "c ! ".join(" ", @entry_tokens[2..$num_entry_tokens-2]);

         if ($debug_msg > 0)
         {
            print "Entry arriv_rate list <@entry_tokens[1]>  $entry_list_COV{@entry_tokens[1]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "f")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify phase type.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_phase_parameters(2, \@entry_tokens, \&check_valid_integer);

         $entry_list_phasetype{@entry_tokens[1]} = "f ! ".join(" ", @entry_tokens[2..$num_entry_tokens-2]);

         if ($debug_msg > 0)
         {
            print "Entry arriv_rate list <@entry_tokens[1]>  $entry_list_phasetype{@entry_tokens[1]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "i")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify fan in.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_identifier_name(@entry_tokens[2]); 
         &check_entry_existance(@entry_tokens[2]);
         &check_valid_phase_parameters(3, \@entry_tokens, \&check_valid_integer);

         if (defined $call_list_fanin{@entry_tokens[1]}{@entry_tokens[2]})
         {
	    die "ERROR at line $cur_line: The FANIN parameters for call from entry: @entry_tokens[1] to entry: @entry_tokens[2] has already been defined.\n";
         }

         $call_list_fanin{@entry_tokens[1]}{@entry_tokens[2]} = "i ! @entry_tokens[2] ".join(" ", @entry_tokens[3..$num_entry_tokens-2]);

         if ($debug_msg > 0)
         {
            print "Call list fanin <@entry_tokens[1]>  $call_list_fanin{@entry_tokens[1]}{@entry_tokens[2]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "o")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify fan out.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_identifier_name(@entry_tokens[2]); 
         &check_entry_existance(@entry_tokens[2]);
         &check_valid_phase_parameters(3, \@entry_tokens, \&check_valid_integer);

         if (defined $call_list_fanout{@entry_tokens[1]}{@entry_tokens[2]})
         {
	    die "ERROR at line $cur_line: The FANOUT parameters for call from entry: @entry_tokens[1] to entry: @entry_tokens[2] has already been defined.\n";
         }

         $call_list_fanout{@entry_tokens[1]}{@entry_tokens[2]} = "o ! @entry_tokens[2] ".join(" ", @entry_tokens[3..$num_entry_tokens-2]);

         if ($debug_msg > 0)
         {
            print "Call list fanout <@entry_tokens[1]> $call_list_fanout{@entry_tokens[1]}{@entry_tokens[2]}\n";
         }

      }
      elsif (@entry_tokens[0] eq "p")
      {
         if ($num_entry_tokens < 3)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify entry priority.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_integer(@entry_tokens[2]);

         $entry_list_priority{@entry_tokens[1]} = @entry_tokens[2];

         if ($debug_msg > 0)
         {
            print "Entry list priority <@entry_tokens[1]>  $entry_list_priority{@entry_tokens[1]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "s")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify entry service time.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_phase_parameters(2, \@entry_tokens, \&check_valid_real);

         $entry_list_service{@entry_tokens[1]} = "s ! ".join(" ", @entry_tokens[2..$num_entry_tokens-2]);

#         &add_SPEX_observation_variables(\@SPEX_observations);

         if (($op_mode eq "SPEX") && ($#SPEX_observations != -1))
         {
            &add_SPEX_observation_variables(\@SPEX_observations);

	 # This $cur_entry is a reminisent of the spreadconv code, get rid of it!!!!!!!!!
	    $cur_entry = @entry_tokens[1];
	    check_SPEX_observations(\@SPEX_observations, "entry");

	    $entry_list_service{@entry_tokens[1]} = $entry_list_service{@entry_tokens[1]}.";";

	 # tack on the SPEX observation keys:

            foreach $cur_obs (@SPEX_observations)
            {
               $entry_list_service{@entry_tokens[1]} = $entry_list_service{@entry_tokens[1]}."%".$cur_obs." ";
            }
	 }

         if ($debug_msg > 0)
         {
            print "Entry list service time  <@entry_tokens[1]>  $entry_list_service{@entry_tokens[1]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "y")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify a synchronous call.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_identifier_name(@entry_tokens[2]); 
         &check_entry_existance(@entry_tokens[2]);
         &check_valid_phase_parameters(3, \@entry_tokens, \&check_valid_real);

         if (defined $call_list_number{@entry_tokens[1]}{@entry_tokens[2]})
         {
	    die "ERROR at line $cur_line: The CALL parameters for call from entry: @entry_tokens[1] to entry: @entry_tokens[2] has already been defined.\n";
         }

         $call_list_number{@entry_tokens[1]}{@entry_tokens[2]} = "y ! @entry_tokens[2] ".join(" ", @entry_tokens[3..$num_entry_tokens-2]);

         if (($op_mode eq "SPEX") && ($#SPEX_observations != -1))
         {
            &add_SPEX_observation_variables(\@SPEX_observations);

	 # This $cur_entry is a reminisent of the spreadconv code, get rid of it!!!!!!!!!
	    $cur_call = @entry_tokens[1];
	    check_SPEX_observations(\@SPEX_observations, "call");

	    $call_list_number{@entry_tokens[1]}{@entry_tokens[2]} = $call_list_number{@entry_tokens[1]}{@entry_tokens[2]}.";";

	 # tack on the SPEX observation keys:

            foreach $cur_obs (@SPEX_observations)
            {
               $call_list_number{@entry_tokens[1]}{@entry_tokens[2]} = $call_list_number{@entry_tokens[1]}{@entry_tokens[2]}."%".$cur_obs." ";
            }
	 }

         if ($debug_msg > 0)
         {
            print "call list number list (SYNC) <@entry_tokens[1]>  $call_list_number{@entry_tokens[1]}{@entry_tokens[2]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "z")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify an asynchronous call.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_identifier_name(@entry_tokens[2]); 
         &check_entry_existance(@entry_tokens[2]);
         &check_valid_phase_parameters(3, \@entry_tokens, \&check_valid_real);

         if (defined $call_list_fanout{@entry_tokens[1]}{@entry_tokens[2]})
         {
	    die "ERROR at line $cur_line: The CALL parameters for call from entry: @entry_tokens[1] to entry: @entry_tokens[2] has already been defined.\n";
         }

         $call_list_number{@entry_tokens[1]}{@entry_tokens[2]} = "z ! @entry_tokens[2] ".join(" ", @entry_tokens[3..$num_entry_tokens-2]);

         if (($op_mode eq "SPEX") && ($#SPEX_observations != -1))
         {
            &add_SPEX_observation_variables(\@SPEX_observations);
 
	 # This $cur_entry is a reminisent of the spreadconv code, get rid of it!!!!!!!!!
	    $cur_call = @entry_tokens[1];
	    check_SPEX_observations(\@SPEX_observations, "call");

	    $call_list_number{@entry_tokens[1]}{@entry_tokens[2]} = $call_list_number{@entry_tokens[1]}{@entry_tokens[2]}.";";

	 # tack on the SPEX observation keys:

            foreach $cur_obs (@SPEX_observations)
            {
               $call_list_number{@entry_tokens[1]}{@entry_tokens[2]} = $call_list_number{@entry_tokens[1]}{@entry_tokens[2]}."%".$cur_obs." ";
            }
	 }

         if ($debug_msg > 0)
         {
            print "call list number list (ASYNC)  <@entry_tokens[1]>  $call_list_number{@entry_tokens[1]}{@entry_tokens[2]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "A")
      {
         if ($num_entry_tokens < 3)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify binding an activity to an entry.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_identifier_name(@entry_tokens[2]); 

         if (defined $entry_list_activities{@activity_tokens[1]})
         {
            die "ERROR at line $cur_line - An activity has already been bound to entry @activity_tokens[2].\n";
         }

# Here we assume that service time is always specified for an entry in a lqn file...

         if (defined $entry_list_service{@activity_tokens[1]})
         {
            die "ERROR at line $cur_line - An activity cannot be bound to a regular defined entry.\n";
         }

         $entry_list_activities{@entry_tokens[1]} = @entry_tokens[2];
         $activity_list_entry{@entry_tokens[2]} = @entry_tokens[1];

         if ($debug_msg > 0)
         {
            print "entry activitiy list <@entry_tokens[1]>  $entry_list_activities{@entry_tokens[1]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "M")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify Max Service time.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_phase_parameters(2, \@entry_tokens, \&check_valid_real);

         $entry_list_maxservice{@entry_tokens[1]} = "M ! ".join(" ", @entry_tokens[2..$num_entry_tokens-2]);

#         &add_SPEX_observation_variables(\@SPEX_observations);

         if (($op_mode eq "SPEX") && ($#SPEX_observations != -1))
         {
            &add_SPEX_observation_variables(\@SPEX_observations);

	 # This $cur_entry is a reminisent of the spreadconv code, get rid of it!!!!!!!!!
	    $cur_entry = @entry_tokens[1];
	    check_SPEX_observations(\@SPEX_observations, "entry");

	    $entry_list_maxservice{@entry_tokens[1]} = $entry_list_maxservice{@entry_tokens[1]}.";";

	 # tack on the SPEX observation keys:

            foreach $cur_obs (@SPEX_observations)
            {
               $entry_list_maxservice{@entry_tokens[1]} = $entry_list_maxservice{@entry_tokens[1]}."%".$cur_obs." ";
            }
	 }

         if ($debug_msg > 0)
         {
            print "Entry list max service time  <@entry_tokens[1]>  $entry_list_maxservice{@entry_tokens[1]}\n";
         }

#         die "ERROR at line $cur_line - Max service time is not supported by this tool, or LQNview or LQNdef.\n";
      }
      elsif (@entry_tokens[0] eq "F")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify a forwarding request.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_identifier_name(@entry_tokens[2]); 
         &check_entry_existance(@entry_tokens[2]);
         &check_valid_real(@entry_tokens[3]);

         if ((@entry_tokens[3] > 1) || (@entry_tokens[3] < 0))
         {
            die "ERROR at line $cur_line - Forwarding probability must be between 0 and 1.\n";
         }

         if (defined $call_list_forwarding{@entry_tokens[1]}{@entry_tokens[2]})
         {
	    die "ERROR at line $cur_line: The FORWARDING parameters for call from entry: @entry_tokens[1] to entry: @entry_tokens[2] has already been defined.\n";
         }

         $call_list_forwarding{@entry_tokens[1]}{@entry_tokens[2]} = "F ! @entry_tokens[2] @entry_tokens[3]";

         if ($debug_msg > 0)
         {
            print "call list number (FORWARDING) <@entry_tokens[1]>  $call_list_forwarding{@entry_tokens[1]}{@entry_tokens[2]}\n";
         }
      }
      elsif (@entry_tokens[0] eq "Z")
      {
         if ($num_entry_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify entry think time.\n";
         }

         &check_valid_identifier_name(@entry_tokens[1]); 
         &check_entry_existance(@entry_tokens[1]);
         &check_valid_phase_parameters(2, \@entry_tokens, \&check_valid_real);

         $entry_list_thinktime{@entry_tokens[1]} = "Z ! ".join(" ", @entry_tokens[2..$num_entry_tokens-2]);

         if ($debug_msg > 0)
         {
            print "entry list thinktime <@entry_tokens[1]>  $entry_list_thinktime{@entry_tokens[1]}\n";
         }
      }
      else
      {
         die "ERROR on line $cur_line - specifying an invalid entry declaration field \"@entry_tokens[0]\" in ENTRY BLOCK.\n";
      }

      @entry_tokens = &get_next_line(0, \@SPEX_observations);
      $num_entry_tokens = @entry_tokens;
   }
}


sub parse_activity_block
{
   my @SPEX_observations = ();
   my $task_name = $_[0];

   @activity_tokens = &get_next_line(0, \@SPEX_observations);
   $num_activity_tokens = @activity_tokens;
   my $stop_parsing = 0;
   my $block_finished = 0;

   while (1)
   {
      if (@activity_tokens[0] eq "c")
      {
         if ($num_activity_tokens < 3)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify coefficient of variation.\n";
         }

         &check_valid_identifier_name(@activity_tokens[1]); 
         &check_valid_real(@activity_tokens[2]);

         $activity_list_COV{@activity_tokens[1]} = @activity_tokens[2];

         if ($debug_msg > 0)
         {
            print "Activity COV list <@activity_tokens[1]>  $activity_list_COV{@activity_tokens[1]}\n";
         }
      }
      elsif (@activity_tokens[0] eq "f")
      {
         if ($num_activity_tokens < 3)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify phase type.\n";
         }

         &check_valid_identifier_name(@activity_tokens[1]); 
         &check_valid_real(@activity_tokens[2]);

         $activity_list_phasetype{@activity_tokens[1]} = @activity_tokens[2];

         if ($debug_msg > 0)
         {
            print "Activity phasetype list <@activity_tokens[1]>  $activity_list_phasetype{@activity_tokens[1]}\n";
         }
      }
      elsif (@activity_tokens[0] eq "s")
      {
         if ($num_activity_tokens < 3)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify activity service time.\n";
         }

         &check_valid_identifier_name(@activity_tokens[1]); 
         &check_valid_real(@activity_tokens[2]);

         $activity_list_service{@activity_tokens[1]} = @activity_tokens[2];

         if ($debug_msg > 0)
         {
            print "Activity service time list  <@activity_tokens[1]>  $activity_list_service{@activity_tokens[1]}\n";
         }
      }
      elsif (@activity_tokens[0] eq "y")
      {
         if ($num_activity_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify a synchronous call.\n";
         }

         if ($#SPEX_observations > 0)
         {
            die "ERROR at line $cur_line - SPEX observations are not allowed in activity calls.\n";
         }

         &check_valid_identifier_name(@activity_tokens[1]); 
         &check_valid_identifier_name(@activity_tokens[2]); 
         &check_valid_real(@activity_tokens[3]);

         $activity_calls{@activity_tokens[1]}{@activity_tokens[2]} = "y ! @activity_tokens[2] @activity_tokens[3]";

         if ($debug_msg > 0)
         {
            print "Activity calls (SYNC) <@activity_tokens[1]>  $activity_calls{@activity_tokens[1]}{@activity_tokens[2]}\n";
         }
      }
      elsif (@activity_tokens[0] eq "z")
      {
         if ($num_activity_tokens < 4)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify an asynchronous call.\n";
         }

         if ($#SPEX_observations > 0)
         {
            die "ERROR at line $cur_line - SPEX observations are not allowed in activity calls.\n";
         }

         &check_valid_identifier_name(@activity_tokens[1]); 
         &check_valid_identifier_name(@activity_tokens[2]); 
         &check_valid_real(@activity_tokens[3]);

         $activity_calls{@activity_tokens[1]}{@activity_tokens[2]} = "z ! @activity_tokens[2] @activity_tokens[3]";

         if ($debug_msg > 0)
         {
            print "Activity call list (ASYNC)  <@activity_tokens[1]>  $activity_calls{@activity_tokens[1]}{@activity_tokens[2]}\n";
         }
      }
      elsif (@activity_tokens[0] eq "Z")
      {
         if ($num_activity_tokens < 3)
         {
            die "ERROR at line $cur_line: There are not enough parameters to specify activity think time.\n";
         }

         &check_valid_identifier_name(@activity_tokens[1]); 
         &check_valid_real(@activity_tokens[2]);

         $activity_list_thinktime{@activity_tokens[1]} = @activity_tokens[2];

         if ($debug_msg > 0)
         {
            print "Activity list thinktime <@activity_tokens[1]>  $activity_list_thinktime{@activity_tokens[1]}\n";
         }
      }
      elsif (@activity_tokens[0] eq "-1")
      {
	 $block_finished = 1;
         last;
      }
      else
      {
         die "ERROR on line $cur_line - specifying an invalid entry declaration field \"@activity_tokens[0]\" in ACTIVITY BLOCK.\n";
      }

      if (defined $task_activity_lookup_list{$task_name})
      {
         if (index($task_activity_lookup_list{$task_name}, @activity_tokens[1].":") == -1)
         {
            $task_activity_lookup_list{$task_name} = $task_activity_lookup_list{$task_name}.@activity_tokens[1].":";
	 }
      }
      else
      {
         $task_activity_lookup_list{$task_name} = @activity_tokens[1].":";
      }

      if ($stop_parsing == 1)
      {
         last;
      }
     
      @activity_tokens = &get_next_line(0, \@SPEX_observations);
      $num_activity_tokens = @activity_tokens;

      $temp_check = join(" ", @activity_tokens);
      
      if (index($temp_check, ":") != -1)
      {
         if ($#SPEX_observations > 0)
         {
            die "ERROR at line $cur_line - SPEX observations are not allowed at the seperator (:) of activity definition and activity connection.\n";
         }

         @activity_connection_half = split(":", $temp_check);
         @activity_tokens = split(" ", @activity_connection_half[0]);
         $num_activity_tokens = @activity_tokens;

         $stop_parsing = 1;

         if (@activity_connection_half[1] =~ /.+/)
         {
            $act_connections = @activity_connection_half[1]."\n";
         }

         if ($num_activity_tokens == 0)
         {
            last;
         }
      } 
   }

   if ($block_finished == 1)
   {
      return;
   }

# now we read in the activity connections till we hit -1 (end list)
# check how robust this stopping condition below is:

   my $last_line = get_next_line_until(0, "-1", \$act_connections);

   $activity_connections{$task_name} = $act_connections;

   if ($debug_msg > 0)
   {
      print "Activity connections for task \"$task_name\" are:\n$activity_connections{$task_name}\n";
   }
   $act_connections = "";
}

sub check_valid_phase_parameters
{
# $_[0] : Initial Offset in the array to start look at
# $_[1] : Actual array we will be examining
# $_[2] : The function to be called to verify the parameter's correctness

   $x = $_[0];

   while ((@{$_[1]}[$x] ne "-1") && (@{$_[1]}[$x] ne ""))
   {
#      print "\n {$_[1]}[$x] '$x'";
      &{$_[2]}(@{$_[1]}[$x]);
      $x++;
   }

   if (@{$_[1]}[$x] eq "")
   {
      die "ERROR on line $cur_line: This line has to be terminated with a -1 to indicate the end of the phase parameters.\n";
   }
}


sub check_entry_existance
{
# $_[0]: The entry name that will be checked to see if it exists

   @tasks = keys %task_list1;
   $entry_found = 0;

   foreach $cur_task (@tasks)
   {
# Strip out the SPEX variables so there is no confusion between entry names and SPEX variables!!!
# Just in case someone wants to use entry names as SPEX variables...

      $task_entry = $task_list1{$cur_task};
      $task_entry =~ s/\$[a-zA-Z_0-9\-]+//g;

      if (index($task_entry, $_[0]) != -1)
      {
         $entry_found = 1;
         last;
      }
   }

   if ($entry_found == 0)
   {
      die "ERROR at line $cur_line - Entry $_[0] has been defined on this line, but not in the task block.\n"; 
   }
}


sub parse_activity_connection
{
# $_[0] contains the current task
# $_[1] contains the activity connection syntax

   if ($debug_msg > 0)
   {
      print "Found an Activity Connection with the following syntax: $_[1]\n";
   }

   if (defined $activity_connections{$_[0]})
   {
      $activity_connections{$_[0]} = $activity_connections{$_[0]}."\n".$_[1];
   }
   else
   {
      $activity_connections{$_[0]} = $_[1];
   }
}



sub add_SPEX_observation_variables
{
# $_[0] is an array of SPEX observation keys variables, and this function extracts the variables and adds them to a global list

   my $num_SPEX_obs = @{$_[0]};
   my @SPEX_obs_vars = ();

   if ($num_SPEX_obs != 0)
   {
# We re-construct the SPEX observation keys into a single line that basically looks like something in the input file

      $SPEX_obs_vars_line = join(" ", @{$_[0]});

# Then extract the variables from this line

      &extract_SPEX_variables(\$SPEX_obs_vars_line, \@SPEX_obs_vars);

# Then concatenate it to a master string de-limited by ",".  Easier to do this than adding to an array...

      if ($SPEX_obs_var_list eq "")
      {
         $SPEX_obs_var_list = join(",", @SPEX_obs_vars);
      }
      else
      {
         $SPEX_obs_var_list = $SPEX_obs_var_list.",".join(",", @SPEX_obs_vars);
      }
   }
}

# This will find the next non-blank line in the file.

sub get_next_line
{
# $_[0]: 1 if it possible to get an EOF
#        0 if it is not possible to get an EOF

# $_[1]: A reference to a variable passed only if we are in SPEX mode, and 
#        can possibly be expecting SPEX observational parameters.

   my $num_params = @_;
   my @control_vars;
   my $line;

   $exit_on_error = 1;

   while (chop($line = <$input_file_handle>))
   {
# replace all tabs characters with spaces

      $line =~ s/\t/ /g;

# strip out the CTRL-M's if they exist.  DOS/Windows like to put a CTRL-M at the end of each line

      $line =~ s/\015//;
       
# check to see if the beginning of the line starts with a space, if so strip off all spaces at the beginning of $line.
      if (index($line, " ") == 0)
      {
         $line =~ s/ +//;
      }

      $cur_line++;

# Erase any quotation marks, Excel puts them in when you have characters . or , in the cell

#     $line =~ s/\"//g;

# Erase all comments

     $line =~ s/\s*#.*$//g;

# If we are in SPEX mode, check to see if we have to look out for SPEX observational parameters

      if ($num_params == 2)
      {
# Here we split off the SPEX observational data from the normal LQN information using the % as a de-limiter

         @temp = split("%", $line);
         $num_temp = @temp;

# Clear out the variable where we will put the SPEX observations into

         @{$_[1]} = ();

         if ($num_temp > 1)
         {
            if ($op_mode eq "SPEX")
            { 
               @{$_[1]} = @temp[1 .. ($num_temp-1)];
               $line = @temp[0];
	    }
            else
            {
               die "ERROR at line $cur_line: There is a SPEX observation key present when the parser is in LQN mode.\n";
	    }
	 }
         else
         {
	    @temp[0] = $line;
	 }

# Now lets look for model control variables

         &extract_SPEX_variables(\@temp[0], \@control_vars);

         $num_control_vars = @control_vars;

         if ($num_control_vars > 0)
         {
            if ($op_mode eq "SPEX")
            {
               if ($SPEX_control_var_list eq "")
               {
                  $SPEX_control_var_list = join(",", @control_vars);
               }
               else
               {
                  $SPEX_control_var_list = $SPEX_control_var_list.",".join(",", @control_vars);
               }
	    }
            else
	    {
               die "ERROR at line $cur_line: There is a SPEX variable present when the parser is in LQN mode.\n";
	    }
	 }
      }

      @line_tokens = split(/ +/, $line);

      $num_line_tokens = @line_tokens;

      if ($num_line_tokens != 0)
      {
         $exit_on_error = 0;
         last;
      }
   }

   if ($exit_on_error == 1)
   {
      if ($_[0] == 0)
      {
         die "ERROR at line $cur_line: Unexpected end of file\n";
      }
      else
      {
         @line_tokens = @line_dummy;
      }
   }
   
   @line_tokens;
}


sub get_next_line_until
{
# $_[0] 1 -> if it is possible to get EOF, 0 -> should not be getting an EOF
# $_[1] stop until we see this string in the current line
# $_[2] variable being passed in by reference that we collect information into until stopping condition is met

   my $no_EOF = 0;
   my $line;
   my @temp_line = ();

   while ($line = <$input_file_handle>)
   {
      $cur_line++;

# replace all tabs characters with spaces

      $line =~ s/\t/ /g;

# strip out the CTRL-M's if they exist.  DOS/Windows like to put a CTRL-M at the end of each line

      $line =~ s/\015//;

# Erase any quotation marks, Excel puts them in when you have characters . or , in the cell

#      $line =~ s/\"//g;

# Erase all comments

#      $line =~ s/\s*#.*$//g;
#      chop($line);

      if ($line =~ /$_[1]/)
      {
         $no_EOF = 1;
         last;
      }
 
      ${$_[2]} = ${$_[2]}.$line;
   }

   if (($no_EOF == 0) && ($_[0] == 0))
   {
      die "ERROR - Could not find keyword $_[1] in the input file.\n";
   }
 # return comment stripped line
   $line
}

1;
