use Getopt::Std;

require parse_check;
require parse_generate;
require parse_parse;

########### Global Definitions:

$debug_msg = 0;                     # If 1 then print debugging messages
$disable_cycle_checking = 0;        # If 1 then disable cycle checking
$enable_SPEX_control_dupe_check = 0;# If 1 then enable SPEX control variable dupe checking
$disable_model_error_checking = 0;  # If 1 then disables error check on the model
$revision = "1.0";                  # Revision number
$input_file = "none";               # filename of the input file
$output_file = "none";              # filename of the output file

########### State information

$defined_SPEX_report = 0;
$defined_SPEX_convergence = 0;
$cur_model_parse_state = "none";
$verify_mode = 0;

$cur_line = 0;
$cur_processor = "none";
$cur_task = "none";
$cur_entry = "none";
$cur_activity = "none";
$cur_number_of_phases = -1;
$op_mode = "none";


########### Variables that contain data retrieved from input file

$model_comments = "";
$model_pragmas = "";
$iteration_limit;
$convergence_limit="";
$print_interval;
$underrelaxation_coeff;

$SPEX_init_block = "";
$SPEX_report_block = "";
$SPEX_convergence_block = "";

$SPEX_obs_var_list = "";
$SPEX_control_var_list ="";

%processor_list = ();
%task_list1 = ();
%task_list2 = (); 

%entry_list_priority = ();
%entry_list_arriv_rate = ();
%entry_list_service = ();
%entry_list_COV = ();
%entry_list_maxservice = ();
%entry_list_phasetype = ();
%entry_list_thinktime = ();
%entry_list_activities = ();

%call_list_number = ();
%call_list_fanin = ();
%call_list_fanout = ();
%call_list_forwarding = ();

%activity_list_service = ();
%activity_list_COV = ();
%activity_list_phasetype = ();
%activity_list_thinktime = ();
%activity_connections = ();
%activity_calls = ();
%task_activity_lookup_list = ();

sub start_program
{
#get the command line options before we do anything else

   &get_command_line_options;
   &open_input_file($input_file);
   @tokens = &parse_initial_information;
   $num_tokens = @tokens;

# This is the main loop which parses the model specification part of the text file.

   while ($num_tokens != 0)
   {
      if (@tokens[0] eq "Processor:")
      {
	 if ((@tokens[1] ne "Name") || (@tokens[2]  ne "Discipline") || (@tokens[3] ne "Multiplicity") || (@tokens[4] ne "Quantum") || (@tokens[5] ne "Replicas") || (@tokens[6] ne "Proc_Rate"))
	 {
	    die "ERROR at line $cur_line: template information seems missing or corrupt.";
	 }
	 else
	 {
            if ($num_tokens != 7)
	    {
	       die "ERROR at line $cur_line: There are too many or too few headings present for the PROCESSOR specification.\n";
	    }

	    &advance_model_parsing_state("processor");
	    $cur_processor = &parse_processor_parameters;
	 }
      }
      elsif (@tokens[0] eq "Task:")
      {
	 if ((@tokens[1] ne "Name") || (@tokens[2] ne "Discipline") || (@tokens[3] ne "Multiplicity") || (@tokens[4] ne "Think_Time") || (@tokens[5] ne "Replicas") || (@tokens[6] ne "Priority"))
	 {
	    die "ERROR at line $cur_line: template information seems missing or corrupt.";
	 }
	 else
	 {
            if ($num_tokens != 7)
	    {
		die "ERROR at line $cur_line: There are too many or too few headings present for the TASK specification.\n";
	    }
	    &advance_model_parsing_state("task");
	    $cur_task = &parse_task_parameters($cur_processor);

	    if ($debug_msg > 0)
	    {
	       print "Current task is now set to $cur_task\n";
	    }
	 }
      }
      elsif (@tokens[0] eq "Entry:")
      {
	 if ((@tokens[1] ne "Name") || (@tokens[2] ne "Priority") || (@tokens[3] ne "Arrival_Rate"))
	 {
	    die "ERROR at line $cur_line: template information seems missing or corrupt.\n";
	 }
	 else
	 {
            if ($num_tokens != 4)
	    {
		die "ERROR at line $cur_line: There are too many or too few headings present for the ENTRY specification.\n";
	    }

	    &advance_model_parsing_state("entry");
	    ($cur_entry, $cur_number_of_phases) = &parse_entry_parameters($cur_task);
	 }
      }
      elsif (@tokens[0] eq "Calls:")
      {
         if ($num_tokens == 3)
         {
            if ((@tokens[1] ne "To_Entry_Name") || (@tokens[2] ne "Sync/Async"))
            {
              die "ERROR at line $cur_line - Heading information for Call parameters seems missing or corrupt\n";
            }

            &advance_model_parsing_state("call");
            &parse_entry_call_parameters($cur_number_of_phases, $cur_entry);
	 }
         elsif ($num_tokens == 4)
         {
            if ((@tokens[1] ne "To_Entry_Name") || (@tokens[2] ne "Sync/Async") || (@tokens[3] ne "Number_Of_Calls"))
            {
               die "ERROR at line $cur_line - Heading information for Call parameters seems missing or corrupt\n";
            }

            &advance_model_parsing_state("call");
            &parse_activity_call_parameters($cur_activity);
         }
         else
         {
            die "ERROR at line $cur_line - There seems to be missing or excess amount of Call heanding information present.\n";
         }
      }
      elsif (@tokens[0] eq "Req_Forwarding:")
      {
         if ((@tokens[1] ne "To_Entry_Name") || (@tokens[2] ne "Prob_Forwarding"))
	 {
	    die "ERROR at line $cur_line: template information seems missing or corrupt.\n";
	 }
	 else
	 {
            if ($num_tokens != 3)
	    {
		die "ERROR at line $cur_line: There are too many or too few headings present for the Request_Forwarding heading specification.\n";
	    }

	    &advance_model_parsing_state("request_forwarding");
	    &parse_request_forwarding_parameters($cur_entry, $cur_task);
	 }
      }
      elsif (@tokens[0] eq "Activity:")
      {
         if ((@tokens[1] ne "Name") || (@tokens[2] ne "Service_Time") || (@tokens[3] ne "Think_Time") || (@tokens[4] ne "COV") ||
             (@tokens[5] ne "Phase_Type") || (@tokens[6] ne "Bind_To_Entry"))
	 {
	    die "ERROR at line $cur_line: template information for ACTIVITY seems missing or corrupt.\n";
	 }
	 else
	 {
            if ($num_tokens != 7)
	    {
		die "ERROR at line $cur_line: There are too many or too few headings present for the ACTIVITY specification.\n";
	    }

	    &advance_model_parsing_state("activity");
	    $cur_activity = &parse_activity_parameters($cur_task);
	 }
      }
      elsif (@tokens[0] eq "Activity_Connection:")
      { 
         &advance_model_parsing_state("activity_connection");
	 &parse_activity_connection($cur_task, join(" ", @tokens[1..$#tokens]));
      }
      elsif (@tokens[0] eq "SPEX_REPORT_START")
      { 
	 if ($op_mode ne "SPEX")
	 {
	    die "ERROR - Parser is not in SPEX mode, yet SPEX_REPORT_START is being defined.\n";
	 }

	 &advance_model_parsing_state("end_model");
	 $defined_SPEX_report = 1;
	 &get_next_line_until(0, "SPEX_REPORT_END", \$SPEX_report_block);
      }
      elsif (@tokens[0] eq "SPEX_CONVERGENCE_START")
      {
	 if ($op_mode ne "SPEX")
	 {
	    die "ERROR - Parser is not in SPEX mode, yet SPEX_REPORT_START is being defined.\n";
	 }

	 &advance_model_parsing_state("end_model");
	 $defined_SPEX_convergence = 1;
	 &get_next_line_until(0, "SPEX_CONVERGENCE_END", \$SPEX_convergence_block);
      }

      @tokens = &get_next_line(1);
      $num_tokens = @tokens;
   }

   &close_input_file;
  
   if ($op_mode eq "SPEX")
   {
      if ($defined_SPEX_report == 0)
      {
	 die "ERROR - SPEX report block has not been defined.\n";
      }
   }

   if ($disable_model_error_checking == 0)
   {
      &check_model_for_errors;
   }
   else
   {
      print "Checking model for errors has been DISABLED\n";
   }

   if ($verify_mode == 0)
   {
     if ($op_mode eq "LQN")
      { 
         &generate_LQN_file($output_file);
      }
      else
      {
         &generate_SPEX_file($output_file);
      }
   }
   else
   {
# If we got this far, then there are no errors in the model file...
       print "No errors found in model.\n";
   }
}

#######################################################

sub get_command_line_options
{
  my @temp;
  my @temp2;
  my $temp;

  $opt_d = 0;
  $opt_h = 0;
  $opt_V = 0;
  $opt_v = 0;
  $opt_e = 0;
  
  getopts('VvdcDeh');

  if ($opt_d)
  {
     $debug_msg = 1;
  }

  if ($opt_h)
  {
     print "Usage: spreadconv [-VvdcDeh] input_file [output_file]\n\n";
     print "-V           displays revision number of program.\n";
     print "-v           just verify model and do not generate a file.\n";
     print "-d           displays debugging output while parsing the spreadsheet input file.\n";
     print "-c           disables cycle checking for synchronous calls (note: when Activities are used\n";
     print "             there is one case where the cycle checker will detect a cycle even though\n";
     print "             it may not exist)\n";
     print "-D           enables checking for duplicate SPEX control variables inside the model.\n";
     print "-e           disables model error checking.\n";
     print "-h           displays this help message.\n";
     print "input_file   ASCII text output from a spreadsheet in the correct format.\n";
     print "output_file  filename of the LQN or SPEX output file.\n\n";
     exit(1);
  }

  if ($opt_V)
  {
     print "Spreadsheet to SPEX/LQN Interface Parser, Version $revision\n";
     print "Copyright 1998, Real-time and Distributed Systems Group\n";
     print "Department of Systems and Computer Engineering,\n";
     print "Carleton University, Ottawa, Canada.\n\n";
     exit(1);
  }

  if ($opt_v)
  {
     $verify_mode = 1;
  }

  if ($opt_c)
  {
     $disable_cycle_checking = 1;
  }

  if ($opt_D)
  {
    $enable_SPEX_control_dupe_check = 1;
  }

  if ($opt_e)
  {
     $disable_model_error_checking = 1;
  }

  $input_file = shift @ARGV;

  if ($input_file eq "")
  {
     die "ERROR - you must specify at least an input filename.  Use -h for more information on usage of this program.\n";
  }

  if ($#ARGV > -1)
  {
     $temp = shift @ARGV;
     @temp2 = split(/\./, $temp);
     $output_file = @temp2[0];
  }
  else
  {
     @temp = split(/\./, $input_file);
     $output_file = @temp[0];
  }

# At this point $#ARGV is going to be -1...
}



########################################################

sub advance_model_parsing_state
{
# $_[0] the next state the model parser wants to go into

   my $next_state = $_[0];
   my $bad_transition = 0;

   if ($cur_model_parse_state eq "none")
   {
      if (($next_state ne "processor") && ($next_state ne "end_model"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "processor")
   {
      if ($next_state ne "task")
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "task")
   {
      if (($next_state ne "entry") && ($next_state ne "activity") && ($next_state ne "activity_connection"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "entry")
   {
      if (($next_state ne "task") && ($next_state ne "entry") && ($next_state ne "call") && ($next_state ne "processor") && ($next_state ne "end_model")
           && ($next_state ne "activity") && ($next_state ne "activity_connection") && ($next_state ne "request_forwarding"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "call")
   {
      if (($next_state ne "call") && ($next_state ne "processor") && ($next_state ne "entry") && ($next_state ne "task") && ($next_state ne "end_model")
         && ($next_state ne "activity") && ($next_state ne "activity_connection"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "activity")
   {
      if (($next_state ne "call") && ($next_state ne "processor") && ($next_state ne "entry") && ($next_state ne "task") && ($next_state ne "end_model")
         && ($next_state ne "activity") && ($next_state ne "activity_connection"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "activity_connection")
   {
      if (($next_state ne "processor") && ($next_state ne "entry") && ($next_state ne "task") && ($next_state ne "end_model")
         && ($next_state ne "activity") && ($next_state ne "activity_connection"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "request_forwarding")
   {
      if (($next_state ne "processor") && ($next_state ne "entry") && ($next_state ne "task") && ($next_state ne "end_model")
         && ($next_state ne "call") && ($next_state ne "activity") && ($next_state ne "activity_connection"))
      {
         $bad_transition = 1;
      }
   }
   elsif ($cur_model_parse_state eq "end_model")
   {
      if ($next_state ne "end_model")
      {
	  die "ERROR - The end of the model has been specified by keyword SPEX_REPORT_START or SPEX_CONVERGENCE_START.  Model definition cannot continue after this point.\n";
      }
   }
   else
   {
      die "INTERNAL ERROR - Model parser state machine has gone into an invalid state: $cur_model_parse_state.\n";
   }

   if ($bad_transition == 0)
   {
      $cur_model_parse_state = $next_state;
   }
   else
   {
      if ($next_state eq "end_model")
      { 
         die "ERROR at line: $cur_line: The model definition cannot end with a $cur_model_parse_state block.\n";
      }
      else
      {
         die "ERROR at line: $cur_line: You cannot specify an $next_state block after a $cur_model_parse_state block.\n";
      }
   }
}


###################################################


sub extract_SPEX_variables
{
# $_[0] variable reference to glob of text which contains the SPEX variables
# $_[1] variable reference a blank array which will contain a list of variables extracted from the
#       glob of text.

   my @rough_var_list_temp = ();
   my @rough_var_list = ();
   my $cur_var;

   @rough_var_list_temp = split(/\$/, ${$_[0]});

   my $num_vars = @rough_var_list_temp;

# element #0 is always going to be blank or have data which is not a variable

   @rough_var_list = @rough_var_list_temp[1 .. ($num_vars-1)];
   $z = 0;

   foreach $cur_var (@rough_var_list)
   {
# find the point where the variable will end, and substitute that point with something we know

      $cur_var =~ s/[\n|,| |\[|\]|*|\/|+|=|\-|\;|\(|\)|\{|\}|\^|\%|\:|\.|<|>|\?]/ZZZxxxZZZ/;

# split the variable de-limiting by the special key we insterted, and we have the variable isolated

      my @var_fin = split("ZZZxxxZZZ", $cur_var);

      @{$_[1]}[$z] = @var_fin[0];
      $z++;
   }
}

# We need to return a TRUE (1), otherwise Perl complains.
1;
