require parse_main;
require parse_check;

$input_file_handle;

sub open_input_file
{
# $_[0]: contains the filename of the input file to read

   if (open (INPUT_FILE, $_[0]) != 1)
   {
      die "ERROR - could not open file \"$_[0]\" for reading.\n"; 
   }

   $input_file_handle = *INPUT_FILE;
}


sub close_input_file
{
   close($input_file_handle);
}


sub parse_initial_information
{
   # First line always defines wether this will generate a SPEX or LQN file

   @tokens = get_next_line(0);

   if (@tokens[0] ne "MODE:")
   {
       die "ERROR at line $cur_line: keyword MODE: has to be the first word in the file.\n";
   }

   if ((@tokens[1] ne "LQN") && (@tokens[1] ne "SPEX"))
   {
       die "ERROR at line $cur_line: keyword MODE can only be equal to LQN and SPEX.\n";
   }

   $op_mode = @tokens[1];

   if ($op_mode eq "LQN")
   {
# In LQN mode we ignore everything until the Model Comments part

      @tokens = get_next_line_until(0, "Model_Comments:", \$ignore_text);
   }
   else
   {
# We are in SPEX mode, and we want to get the solver name and parameters
# But first we have to skip over the FILE: and SPEX_PARAMETERS: keyword, as we are not interested in that at all.

      @tokens = get_next_line(0);
      $num_tokens = @tokens;

      if ($num_tokens != 2)
      {
	 die "ERROR at line $cur_line: Something is wrong on the line where FILE: is specified.\n        Too few or too many parameters have been specified.\n";
      }

      if (@tokens[0] ne "FILE:")
      {
         die "ERROR at line $cur_line: Cannot find the keyword FILE:\n";
      }

# skip over the SPEX_PARAMETERS: part...

      @tokens = get_next_line(0);
      $num_tokens = @tokens;

      if (@tokens[0] ne "SPEX_PARAMETERS:")
      {
         die "ERROR at line $cur_line: Cannot find the keyword SPEX_PARAMETERS:\n";
      }

      if ($num_tokens == 1)
      {
         print "WARNING - no parameters are being passed to the SPEX program.\n";
      }

# Get the SOLVER: information

      @tokens = get_next_line(0);
      $num_tokens = @tokens;

      if ($num_tokens != 2)
      {
	 die "ERROR at line $cur_line: Something is wrong on the line where SOLVER: is specified.\n        Too few or too many parameters have been specified.\n";
      }

      if (@tokens[0] ne "SOLVER:")
      {
	 die "ERROR at line $cur_line: The keyword SOLVER: is expected but not found.\n";
      }

      $solver = @tokens[1];

# Get the SOLVER_PARAMETERS: information

      @tokens = get_next_line(0);
      $num_tokens = @tokens;

      if (@tokens[0] ne "SOLVER_PARAMETERS:")
      {
	 die "ERROR at line $cur_line: The keyword SOLVER_PARAMETERS: is expected but not found.\n";
      }

      if ($num_tokens == 1)
      {
	 print "WARNING - No parameters are being passed to the solver.\n";
      }

      $solver_parameters = join(" ", @tokens[1..$num_tokens]);
      @tokens = get_next_line(0);

      $SPEX_init_block = "\$solver = $solver $solver_parameters\n\n";
   }

   # Now we look for the next part, the Model Comments part

# Strip off any spaces...
   @tokens[0] =~ s/\s//g;

   if (@tokens[0] ne "Model_Comments:")
   {
       die "ERROR at line $cur_line: Expecting to find the keyword 'Model_Comments:'.\n";
   }

   get_next_line_until(0, "Pragma:", \$model_comments);

   # Now we scan for the pragmas in a similar manner as with the Model Comments

   $next_line = get_next_line_until(0, "Convergence_Limit:", \$model_pragmas);
#   $model_pragmas ="";

   # If we got this far, then $line = "Convergence_Limit xxx" were xxx is some value
   # NOTE: $line is a global variable that is used by get_next_line_until

# strip off the spaces in the front of the line if they exist

   if (index($next_line, " ") == 0)
   {
      $next_line =~ s/ +//;
   } 

# strip out the CTRL-M's if they exist.  DOS/Windows like to put a CTRL-M at the end of each line

   $next_line =~ s/\015//;

   @tokens = split(/ +/, $next_line);

   $num_tokens = @tokens;

   if ($num_tokens != 2)
   {
       die "ERROR $num_tokens at line $cur_line: Looks like there is a missing parameter for Convergence Limit.\n";
   }

   $convergence_limit = @tokens[1];

   # Now get the Iteration Limit

   @tokens = get_next_line(0);
   $num_tokens = @tokens;

   if (@tokens[0] ne "Iteration_Limit:")
   {
       die "ERROR at $cur_line: Heading information 'Iteration_Limit' seems to be damaged or missing.\n";
   }

   if ($num_tokens != 2)
   {
       die "ERROR at line $cur_line: Looks like there is a missing parameter for Iteration Limit.\n";
   }

   $iteration_limit = @tokens[1];

   # Now get the Print Interval

   @tokens = get_next_line(0);
   $num_tokens = @tokens;

   if (@tokens[0] ne "Print_Interval:")
   {
       die "ERROR at line $cur_line: Heading information 'Print_Interval' seems to be damaged or missing.\n";
   }

   if ($num_tokens != 2)
   {
       die "ERROR at line $cur_line: Looks like there is a missing parameter for Print Interval.\n";
   }

   $print_interval = @tokens[1];

   # Now get the Under Relaxation Coefficient

   @tokens = get_next_line(0);
   $num_tokens = @tokens;

   if (@tokens[0] ne "Underrelaxation_Coefficient:")
   {
       die "ERROR at line $cur_line: Heading information 'Underrelaxation_Coefficient:' seems to be damaged or missing.\n";
   }

   if ($num_tokens != 2)
   {
       die "ERROR at line $cur_line: Looks like there is a missing parameter for Under Relaxation Coefficient.\n";
   }

   $underrelaxation_coeff = @tokens[1];

   # Get the SPEX initialization block, if we are in SPEX mode

   @tokens = &get_next_line(0);
   $num_tokens = @tokens;

   if (@tokens[0] eq "SPEX_INIT_START")
   {
      if ($op_mode eq "SPEX")
      {
	 $defined_SPEX_init = 1;
	 get_next_line_until(0, "SPEX_INIT_END", \$SPEX_init_block);
         @tokens = &get_next_line(0);
      }
      else
      {
	 die "ERROR - Specifying a SPEX initialization block, yet parser is in LQN mode.\n";
      }
   }
   elsif ($op_mode eq "SPEX")
   {
      die "ERROR - No SPEX initialization block defined, yet the parser is in SPEX mode.\n";
   }
   @tokens;
}


sub parse_processor_parameters
{
   my @SPEX_observations = ();
   @proc_tokens = &get_next_line(0, \@SPEX_observations);
   $num_proc_tokens = @proc_tokens;

   if ($num_proc_tokens != 6)
   {
      die "ERROR at line $cur_line - Invalid number of parameters for processor specification.\n";
   }

   &add_SPEX_observation_variables(\@SPEX_observations);

   if ($debug_msg > 0)
   {
      print "Found processor @proc_tokens[0]\n";
   }

   if (defined $processor_list{@proc_tokens[0]})
   {
      die "ERROR at line $cur_line - the PROCESSOR name @proc_tokens[0] has already been used.\n";
   }

   &check_valid_identifier_name(@proc_tokens[0]);
   &check_valid_integer(@proc_tokens[2]);
   &check_valid_real(@proc_tokens[3]);
   &check_valid_integer(@proc_tokens[4]);
   &check_valid_real(@proc_tokens[5]);
   
   $cur_processor = @proc_tokens[0];

   $processor_list{@proc_tokens[0]} = @proc_tokens[1]." ".@proc_tokens[2]." ".@proc_tokens[3]." ".@proc_tokens[4]." ".@proc_tokens[5];

   if ($op_mode eq "SPEX")
   {
      check_SPEX_observations(\@SPEX_observations, "processor");

      $processor_list{@proc_tokens[0]} = $processor_list{@proc_tokens[0]}.";";

# tack on the SPEX observation keys:

      foreach $cur_obs (@SPEX_observations)
      {
         $processor_list{@proc_tokens[0]} = $processor_list{@proc_tokens[0]}."  %".$cur_obs;
      }
   }

   @proc_tokens[0];   # return the current processor we are dealing with
}


sub parse_task_parameters
{
   my @SPEX_observations = ();
   @task_tokens = &get_next_line(0, \@SPEX_observations);
   $num_task_tokens = @task_tokens;

   if ($num_task_tokens != 6)
   {
      die "ERROR at line $cur_line - Invalid number of parameters for TASK specification.\n";
   }

   &add_SPEX_observation_variables(\@SPEX_observations);

   if ($debug_msg > 0)
   {
      print "Found task @task_tokens[0]\n";
   }

   if (defined $task_list1{@task_tokens[0]})
   {
      die "ERROR at line $cur_line - the TASK name @task_tokens[0] has already been used.\n";
   }

   $cur_task = @task_tokens[0];

   &check_valid_identifier_name(@task_tokens[0]);
   &check_valid_integer(@task_tokens[2]);
   &check_valid_real(@task_tokens[3]);
   &check_valid_integer(@task_tokens[4]);
   &check_valid_integer(@task_tokens[5]);

# task_list1 is the first part of the task specification before the list of entries
   $task_list1{@task_tokens[0]} = @task_tokens[1]." ";

# task_list2 is the trailing information after the list of entries
# $_[0] is the procesor this task belongs to

   $task_list2{@task_tokens[0]} = $_[0]." ".@task_tokens[5]." ".@task_tokens[3]." ".@task_tokens[2]." ".@task_tokens[4];

   if ($op_mode eq "SPEX")
   {
      check_SPEX_observations(\@SPEX_observations, "task");

      $task_list2{@task_tokens[0]} = $task_list2{@task_tokens[0]}.";";

# tack on the SPEX observation keys:

      foreach $cur_obs (@SPEX_observations)
      {
         $task_list2{@task_tokens[0]} = $task_list2{@task_tokens[0]}."  %".$cur_obs;
      }
   }

   @task_tokens[0];  # return the task we are currently dealing with
}


sub parse_entry_parameters
{
   my @SPEX_observations = ();
   my $prev_file_pos = 0;
   my $prev_cur_line = 0;
   @entry_tokens = &get_next_line(0, \@SPEX_observations);
   $num_entry_tokens = @entry_tokens;

   if ($num_entry_tokens != 3)
   {
      die "ERROR at line $cur_line - Invalid number of parameters for ENTRY specification.\n";
   }

   &add_SPEX_observation_variables(\@SPEX_observations);

   if ($debug_msg > 0)
   {
      print "Found entry @entry_tokens[0]\n";
   }

   if (defined $entry_list_service{@entry_tokens[0]})
   {
      die "ERROR at line $cur_line - the entry name @entry_tokens[0] has already been used.\n";
   }

   if (defined $entry_list_activities{@entry_tokens[0]})
   {
      die "ERROR at line $cur_line - the entry name @entry_tokens[0] is already being used by an activity bound to an entry.\n";
   }

   &check_valid_identifier_name(@entry_tokens[0]);
   &check_valid_integer(@entry_tokens[1]);
   &check_valid_real(@entry_tokens[2]);

   $entry_list_priority{@entry_tokens[0]} = @entry_tokens[1];
   $entry_list_arriv_rate{@entry_tokens[0]} = @entry_tokens[2];

   $entry_name = @entry_tokens[0];

# Check the SPEX parameters now, so that we have the correct line number reported when an error occurs

   check_SPEX_observations(\@SPEX_observations, "entry");


# associate this entry with the task in $_[0], by putting the name of the entry into $task_list

   $task_list1{$_[0]} = $task_list1{$_[0]}.@entry_tokens[0]." ";

   @entry_tokens = &get_next_line(0);

   if ((@entry_tokens[0] ne "Service_Time") && (@entry_tokens[1] ne "COV") && (@entry_tokens[2] ne "Phase_Type")
        && (@entry_tokens[3] ne "Think_Time") && (@entry_tokens[4] ne "Max_Service"))
   {
      die "ERROR at line $cur_line - Corrupt or missing heading information for specification of service times.\n";
   }
 
# We expect at least 1 phase!!! 
   @entry_tokens = &get_next_line(0);
   $num_entry_tokens = @entry_tokens;
   $exit_phases = 0;
   my $cur_phase = 1;

   my $prev_file_pos = tell($input_file_handle);
   my $prev_cur_line = $cur_line;

   while ($num_entry_tokens != 0)
   {
      if (@entry_tokens[0] ne "Phase_"."$cur_phase")
      {
         seek($input_file_handle, $prev_file_pos, 0);
         $cur_line = $prev_cur_line;
         $cur_phase--;

         last;
      }
      else
      {
         if ($num_entry_tokens != 6)   # Note the Phase_N heading takes up 1 spot
         {
	     die "ERROR at line $cur_line - There are either too many or too few parameters in service time specifications in phase $cur_phase\n";
	 }

         if ($debug_msg > 0)
         {
            print "Found phase $cur_phase; service_time @entry_tokens[1]; COV  @entry_tokens[2]; phase_type @entry_tokens[3]; and think_time @entry_tokens[4]\n";
	 }

         &check_valid_real(@entry_tokens[1]);
         &check_valid_real(@entry_tokens[2]);
         &check_valid_real(@entry_tokens[4]);
         &check_valid_real(@entry_tokens[5]);

# check if user put a control variable in for Phase_Type by checking first letter of the entry ie: searching for "$"

         if (index(@entry_tokens[3], "\$") != 0)
         {
            if ((@entry_tokens[3] ne "S") && (@entry_tokens[3] ne "D"))
            {
               if ($op_mode eq "LQN")
               {
                  die "ERROR at line $cur_line - Must specify S (Stochastic) or D (Deterministic) for Phase Type.\n";  
               }
               else
               {
                  die "ERROR at line $cur_line - Must specify S (Stochastic), D (Deterministic), or a SPEX variable for Phase Type.\n";
               }
            }
         }

         if (defined $entry_list_service{$entry_name})
         {
	    $entry_list_service{$entry_name} = $entry_list_service{$entry_name}." ".@entry_tokens[1];
            $entry_list_COV{$entry_name} = $entry_list_COV{$entry_name}." ".@entry_tokens[2];
            $entry_list_thinktime{$entry_name} = $entry_list_thinktime{$entry_name}." ".@entry_tokens[4];
            $entry_list_maxservice{$entry_name} = $entry_list_maxservice{$entry_name}." ".@entry_tokens[5];

            if (@entry_tokens[3] eq "S")
            {
               $entry_list_phasetype{$entry_name} = $entry_list_phasetype{$entry_name}." 0";
            }
            elsif (@entry_tokens[3] eq "D")
	    {
               $entry_list_phasetype{$entry_name} = $entry_list_phasetype{$entry_name}." 1";
            }
            else
            {
               $entry_list_phasetype{$entry_name} = $entry_list_phasetype{$entry_name}." ".@entry_tokens[3];
	    }
	 }
         else
         {
	    $entry_list_service{$entry_name} = "s ! @entry_tokens[1]";
            $entry_list_COV{$entry_name} = "c ! @entry_tokens[2]";
            $entry_list_thinktime{$entry_name} = "Z ! @entry_tokens[4]";
            $entry_list_maxservice{$entry_name} = "M ! @entry_tokens[5]";
        
            if (@entry_tokens[3] eq "S")
            {
               $entry_list_phasetype{$entry_name} = "f ! 0";
	    }
            elsif (@entry_tokens[3] eq "D")
            {
               $entry_list_phasetype{$entry_name} = "f ! 1";
            }
            else
            {
               $entry_list_phasetype{$entry_name} = "f ! ".@entry_tokens[3];
            }
         }
         $cur_phase++;
      }
   
      $prev_file_pos = tell($input_file_handle);
      $prev_cur_line = $cur_line;
      @entry_tokens = &get_next_line(1);
      $num_entry_tokens = @entry_tokens;
  }

   if ($cur_phase = 0)
   {
      die "ERROR at line $cur_line - You need to specify at least 1 phase when defining an entry.\n";
   }


   if ($op_mode eq "SPEX")
   {
      $entry_list_service{$entry_name} = $entry_list_service{$entry_name}.";";

# tack on the SPEX observation keys:

      foreach $cur_obs (@SPEX_observations)
      {
         $entry_list_service{$entry_name} = $entry_list_service{$entry_name}."  %".$cur_obs;
      }
   }
   ($entry_name, $cur_phase);
}


sub parse_entry_call_parameters
{
   my @SPEX_observations = ();

   my $entry_num_phases = $_[0];
   my $entry_name = $_[1];
   my $num_call_tokens = -1;
   my $cur_phase=1;
   my $dest_entry;

   @SPEX_observations = ();   # Clear observations from previous call, otherwise it will stay there!
 
   @call_tokens = &get_next_line(0, \@SPEX_observations);
   $num_call_tokens = @call_tokens;

   if ($num_call_tokens != 2)
   {
      die "ERROR at line $cur_line - Missing or there are too many entry call paramters\n";
   }

   &add_SPEX_observation_variables(\@SPEX_observations);

   &check_valid_identifier_name(@call_tokens[0]);

   if ($debug_msg > 0)
   {
      print "Entry: $entry_name is calling entry @call_tokens[0]...\n";
   }

   $dest_entry = @call_tokens[0];

   if (defined $call_list_number{$entry_name}{$dest_entry})
   {
      die "ERROR at line $cur_line: Actual call, or fanin/fanout specification has been defined twice for the call between entries $entry_name and $dest_entry.\n";
   }

   $call_list_fanin{$entry_name}{$dest_entry} = "i ! ".@call_tokens[0]." ";
   $call_list_fanout{$entry_name}{$dest_entry} = "o ! ".@call_tokens[0]." ";

   if (@call_tokens[1] eq "S")   # Checking if it's a synchronous call
   {  
      $call_list_number{$entry_name}{$dest_entry} = "y ! ".@call_tokens[0]." ";
   } 
   elsif (@call_tokens[1] eq "A")   # Checking if it's a asynchronous call
   { 
      $call_list_number{$entry_name}{$dest_entry} = "z ! ".@call_tokens[0]." ";
   }
   else
   {
      if ($op_mode eq "LQN")
      {
	 die "ERROR at line $cur_line - You can only specify an S for Synchronous or A for Asynchronous calls in entry: $entry_name.\n";
      }
      else
      {
	 if (index(@call_tokens[1], "\$") == 0)
	 {
	    die "ERROR at line $cur_line - You cannot specify a SPEX control variable call type.  Valid values are S for Synchronous or A for Asynchronous calls.\n";
	 }
	 else
	 {
	    die "ERROR at line $cur_line - You can only specify an S for Synchronous or A for Asynchronous calls in entry: $entry_name.\n";
	 }
      }
   }

# Read in next line, and check the heading of the table which will be following it

   @call_tokens = &get_next_line(0);
   $num_call_tokens = @call_tokens;

   if ((@call_tokens[0] ne "Number_of_Calls") || (@call_tokens[1] ne "Fan_In") || (@call_tokens[2] ne "Fan_Out"))
   {
      die "ERROR at line $cur_line - The heading specifying the number of calls, fan in/out appears to be corrupt.\n";
   }

   if ($num_call_tokens != 3)
   {
      die "ERROR at line $cur+line - There seem to be too few or too many headings for the calls, fan in/out table.\n";
   }

   my $prev_file_pos = tell($input_file_handle);
   my $prev_cur_line = $cur_line;

   @call_tokens = @call_tokens = &get_next_line(0);
   $num_call_tokens = @call_tokens;

   $non_zero_call = 0;
   @entry_phase_types = split(" ", $entry_list_phasetype{$entry_name});

   while ($num_call_tokens != 0)
   {
      if (@call_tokens[0] ne "Phase_"."$cur_phase")
      {
         seek($input_file_handle, $prev_file_pos, 0);
         $cur_line = $prev_cur_line;
         $cur_phase--;

         last;
      }
      else
      {
         if ($num_call_tokens != 4)   # Note the Phase_N heading takes up 1 spot
         {
	     die "ERROR at line $cur_line - There are either too many or too few parameters in the call specification for phase $cur_phase\n";
	 }

         if ($debug_msg > 0)
         {
            print "Found Calling phase $cur_phase; Number of Calls @call_tokens[1]; Fan In  @call_tokens[2]; Fan Out @call_tokens[3]\n";
	 }

# When phasetype is STOCHASTIC, the number of times an entry can make a call could be a real value.

         if (@entry_phase_types[$cur_phase+1] == 1)
         {
            &check_valid_integer(@call_tokens[1]);
 	 }
         else
         {
            &check_valid_real(@call_tokens[1]);
         }

         &check_valid_integer(@call_tokens[2]);
         &check_valid_integer(@call_tokens[3]);

         if ((@call_tokens[1] > 0) || (index(@call_tokens[1], "\$") == 0))
         {
	    $non_zero_call = 1;
         }
 
         $call_list_number{$entry_name}{$dest_entry} = $call_list_number{$entry_name}{$dest_entry}." ".@call_tokens[1];

######## NOTE: Currently LQNS only accepts 1 value for fanin and fanout.  However it should be capable of handling a vector
######## of values.  The below check for $cur_phase should be removed when LQNS gets updated to read in vectors for
######## fanin and fanout.

         if ($cur_phase == 1)
         {
            $call_list_fanin{$entry_name}{$dest_entry} = $call_list_fanin{$entry_name}{$dest_entry}." ".@call_tokens[2];
            $call_list_fanout{$entry_name}{$dest_entry} = $call_list_fanout{$entry_name}{$dest_entry}." ".@call_tokens[3];
	 }

         $cur_phase++;
      }
   
      $prev_file_pos = tell($input_file_handle);
      $prev_cur_line = $cur_line;
      @call_tokens = &get_next_line(1);
      $num_call_tokens = @call_tokens;
   }

   if ($cur_phase = 0)
   {
      die "ERROR at line $cur_line - You need to specify at least 1 phase in when making a call to another entry.\n";
   }

   if ($cur_phase != $entry_num_phases) 
   {
      die "ERROR at line $cur_line - $cur_phase, $entry_num_phases Need to have same number of call parameters as phases.\n";
   }

   if ($non_zero_call == 0)
   {
      die "ERROR at line $cur_line: Entry: $entry_name has a call specified, but number of calls for all phases is 0.\n";
   }

# stick in the SPEX observations keys for this particular call

   if ($op_mode eq "SPEX")
   {
      check_SPEX_observations(\@SPEX_observations, "call");

      $call_list_number{$entry_name}{$dest_entry} = $call_list_number{$entry_name}{$dest_entry}.";";

      foreach $cur_obs (@SPEX_observations)
      {
	 $call_list_number{$entry_name}{$dest_entry} = $call_list_number{$entry_name}{$dest_entry}."  %".$cur_obs;
      }
   }
}



sub parse_request_forwarding_parameters
{
# $_[0] is current entry we are dealing with
# $_[1] is the current task we are dealing with

   my @SPEX_observations = ();
   my $dest_entry;
   @forward_tokens = &get_next_line(0, \@SPEX_observations);
   $num_forward_tokens = @forward_tokens;

   if ($num_forward_tokens != 2)
   {
      die "ERROR at line $cur_line - Invalid number of parameters for Request_Forwarding specification.\n";
   }

   if ($#SPEX_observations > 0)
   {
      die "ERROR at line $cur_line - SPEX observations are not allowed for request forwarding, regardless of being in SPEX or LQN mode.\n";
   }

   if ($debug_msg > 0)
   {
      print "Found entry forwarding from $_[0] to @forward_tokens[0]\n";
   }

   &check_valid_identifier_name(@forward_tokens[0]);

   if ((@forward_tokens[1] != 0) || (index(@forward_tokens[1], "\$") == 0))   # Checking for Forwarding
   {
   # Check to see if the current task is a reference task, because reference tasks can't forward!

      @task_parts = split(" ", $task_list1{$_[1]});

      if (@task_parts[0] eq "r")
      {
         die "ERROR at line $cur_line - Entry $_[0] has a forwarding probabilty of @forward_tokens[1], when it belongs to a reference task.\n";
      }

      $dest_entry = @forward_tokens[0];

   # forwarding probability should be between 0 and 1, or a variable if it's in SPEX mode

      if ((@forward_tokens[1] <= 1) && (@forward_tokens[1] >= 0))
      {
         $call_list_forwarding{$_[0]}{$dest_entry} = "F ! ".@forward_tokens[0]." ".@forward_tokens[1];
      }
      else
      {
         die "ERROR at line $cur_line - Forwarding Probability in entry: $_[0] should be between 0 and 1.\n";
      }
   }
}

sub parse_activity_parameters
{
# $_[0] is the current task we are in

   my @SPEX_observations = ();
   @activity_tokens = &get_next_line(0, \@SPEX_observations);
   $num_activity_tokens = @activity_tokens;

   if ($num_activity_tokens != 6)
   {
      die "ERROR at line $cur_line - Invalid number of parameters for ACTIVITY specification.\n";
   }

   if ($#SPEX_observations > 0)
   {
      die "ERROR at line $cur_line - SPEX observations for ACTIVITIES are not supported yet.\n";
   }

   if ($debug_msg > 0)
   {
      print "Found activity @activity_tokens[0]\n";
   }

   if (defined $activity_list_service{@activity_tokens[0]})
   {
      die "ERROR at line $cur_line - the ACTIVITY name @activity_tokens[0] has already been used.\n";
   }

   &check_valid_identifier_name(@activity_tokens[0]);
   &check_valid_real(@activity_tokens[1]);
   &check_valid_real(@activity_tokens[2]);
   &check_valid_real(@activity_tokens[3]);
   &check_valid_identifier_name(@activity_tokens[5]);

   if (@activity_tokens[4] eq "S")
   {
      @activity_tokens[4] = 0;
   }
   elsif (@activity_tokens[4] eq "D")
   {
      @activity_tokens[4] = 1;
   }
   else
   {
      die "ERROR at line $cur_line: Unknown Phase Type Flag is being used.\n";
   }

   if (@activity_tokens[1] < 0)
   {
      die "ERROR at line $cur_line: Service times for activities must be a positive real number or at least 0.\n";
   }

   $activity_list_service{@activity_tokens[0]} = @activity_tokens[1];
   $activity_list_thinktime{@activity_tokens[0]} = @activity_tokens[2];
   $activity_list_COV{@activity_tokens[0]} = @activity_tokens[3];
   $activity_list_phasetype{@activity_tokens[0]} = @activity_tokens[4];

# For error checking purposes and Activity block generation we keep track of what task the activity belongs in

   if (defined $task_activity_lookup_list{$_[0]})
   {
      $task_activity_lookup_list{$_[0]} = $task_activity_lookup_list{$_[0]}.@activity_tokens[0].":";
   }
   else
   {
      $task_activity_lookup_list{$_[0]} = @activity_tokens[0].":";
   }

   if (@activity_tokens[5] ne "N/A")
   {
      if (defined $entry_list_activities{@activity_tokens[5]})
      {
         die "ERROR at line $cur_line - An activity has already been bound to entry @activity_tokens[5].\n";
      }

      if (defined $entry_list_service{@activity_tokens[5]})
      {
         die "ERROR at line $cur_line - An activity cannot be bound to a regular defined entry.\n";
      }
      $entry_list_activities{@activity_tokens[5]} = @activity_tokens[0];

# We also have to add this to the list of entries in the task block

      $task_list1{$_[0]} = $task_list1{$_[0]}.@activity_tokens[5]." ";
   }

   @activity_tokens[0];  # return the activity we are currently dealing with
}

sub parse_activity_connection
{
# $_[0] contains the current task
# $_[1] contains the activity connection syntax

   if ($debug_msg > 0)
   {
      print "Found an Activity Connection with the following syntax: $_[1]\n";
   }

   if (defined $activity_connections{$_[0]})
   {
      $activity_connections{$_[0]} = $activity_connections{$_[0]}."\n".$_[1];
   }
   else
   {
      $activity_connections{$_[0]} = $_[1];
   }
}

sub parse_activity_call_parameters
{
   my @SPEX_observations = ();
   my $activity_name = $_[0];

   my @call_tokens = &get_next_line(0, \@SPEX_observations);
   my $num_call_tokens = @call_tokens;

   if ($num_call_tokens != 3)
   {
      die "ERROR at line $cur_line - Missing entry call paramters\n";
   }

   if ($#SPEX_observations > 0)
   {
      die "ERROR at line $cur_line - SPEX observations are not allowed in activity calls.\n";
   }

   if ($debug_msg > 0)
   {
      print "Activity: $activity_name is calling entry @call_tokens[0], @call_tokens[2] times, with call type: @call_tokens[1].\n";
   }

   &check_valid_identifier_name(@call_tokens[0]);
   &check_valid_integer(@call_tokens[2]);

# Ensure we have a positive non-zero number of calls to that entry

   if (@call_tokens[2] <= 0)
   {
      die "ERROR at line $cur_line: Activity $activity_name calling Entry @call_tokens[0] must have positive non-zero visits.\n";
   }

   if (defined $activity_calls{$activity_name}{@call_tokens[0]})
   {
      die "ERROR at line $cur_line: The call definition from activity: $activity_name to entry @call_tokens[0] is being defined twice.\n";
   }

   if (@call_tokens[1] eq "S")   # Checking if it's a synchronous call
   {  
      $activity_calls{$activity_name}{@call_tokens[0]} = "y ! ".@call_tokens[0]." ".@call_tokens[2];
   } 
   elsif (@call_tokens[1] eq "A")   # Checking if it's a asynchronous call
   { 
      $activity_calls{$activity_name}{@call_tokens[0]} = "z ! ".@call_tokens[0]." ".@call_tokens[2];
   }
   else
   {
      die "ERROR at line $cur_line: Call_Type (Sync/Async) in Activity call specification is unknown.\n";
   }
}

sub add_SPEX_observation_variables
{
# $_[0] is an array of SPEX observation keys variables, and this function extracts the variables and adds them to a global list

   my $num_SPEX_obs = @{$_[0]};
   my @SPEX_obs_vars = ();

   if ($num_SPEX_obs != 0)
   {
# We re-construct the SPEX observation keys into a single line that basically looks like something in the input file

      $SPEX_obs_vars_line = join(" ", @{$_[0]});

# Then extract the variables from this line

      &extract_SPEX_variables(\$SPEX_obs_vars_line, \@SPEX_obs_vars);

# Then concatenate it to a master string de-limited by ",".  Easier to do this than adding to an array...

      if ($SPEX_obs_var_list eq "")
      {
         $SPEX_obs_var_list = join(",", @SPEX_obs_vars);
      }
      else
      {
         $SPEX_obs_var_list = $SPEX_obs_var_list.",".join(",", @SPEX_obs_vars);
      }
   }
}

# This will find the next non-blank line in the file.

sub get_next_line
{
# $_[0]: 1 if it possible to get an EOF
#        0 if it is not possible to get an EOF

# $_[1]: A reference to a variable passed only if we are in SPEX mode, and 
#        can possibly be expecting SPEX observational parameters.

   my $num_params = @_;
   my @control_vars;
   my $line, $cur_temp, $temp_var;
#   parse_main->import();

   $exit_on_error = 1;

   while (chop($line = <$input_file_handle>))
   {
# replace all tabs characters with spaces

      $line =~ s/\t/ /g;

# strip out the CTRL-M's if they exist.  DOS/Windows like to put a CTRL-M at the end of each line

      $line =~ s/\015//;
       
# check to see if the beginning of the line starts with a space, if so strip off all spaces at the beginning of $line.
      if (index($line, " ") == 0)
      {
         $line =~ s/ +//;
      }

      $cur_line++;

# Erase any quotation marks, Excel puts them in when you have characters . or , in the cell

     $line =~ s/\"//g;

# Erase all comments

     $line =~ s/\s*#.*$//g;

# If we are in SPEX mode, check to see if we have to look out for SPEX observational parameters

      if ($num_params == 2)
      {
# Here we split off the SPEX observational data from the normal LQN information using the % as a de-limiter

         @temp = split("%", $line);
         $num_temp = @temp;

# Clear out the variable where we will put the SPEX observations into

         @{$_[1]} = ();

         if ($num_temp > 1)
         {
            if ($op_mode eq "SPEX")
            { 
# Check to ensure each observation variable has a $ infront of it to be a valid SPEX variable

               foreach $cur_temp (@temp[1..($num_temp-1)])
               {
	          @temp_var = split(" ", $cur_temp);
                  if (@temp_var[1] !~ /^\$.*/)
                  {
		      die "ERROR on line $cur_line - The variable \"@temp_var[1]\" should have a \$ infront of it.\n";
		  }
	       }
               @{$_[1]} = @temp[1 .. ($num_temp-1)];
               $line = @temp[0];
	    }
            else
            {
               die "ERROR at line $cur_line: There is a SPEX observation key present when the parser is in LQN mode.\n";
	    }
	 }
         else
         {
	    @temp[0] = $line;
	 }

# Now lets look for model control variables

         &extract_SPEX_variables(\@temp[0], \@control_vars);

         $num_control_vars = @control_vars;

         if ($num_control_vars > 0)
         {
            if ($op_mode eq "SPEX")
            {
               if ($SPEX_control_var_list eq "")
               {
                  $SPEX_control_var_list = join(",", @control_vars);
               }
               else
               {
                  $SPEX_control_var_list = $SPEX_control_var_list.",".join(",", @control_vars);
               }
	    }
            else
	    {
               die "ERROR at line $cur_line: There is a SPEX variable present when the parser is in LQN mode.\n";
	    }
	 }
      }

      @line_tokens = split(/ +/, $line);

      $num_line_tokens = @line_tokens;

      if ($num_line_tokens != 0)
      {
         $exit_on_error = 0;
         last;
      }
   }

   if ($exit_on_error == 1)
   {
      if ($_[0] == 0)
      {
         die "ERROR at line $cur_line: Unexpected end of file\n";
      }
      else
      {
         @line_tokens = @line_dummy;
      }
   }
   
   @line_tokens;
}


sub get_next_line_until
{
# $_[0] 1 -> if it is possible to get EOF, 0 -> should not be getting an EOF
# $_[1] stop until we see this string in the current line
# $_[2] variable being passed in by reference that we collect information into until stopping condition is met

   my $no_EOF = 0;
   my $line;
#   parse_main->import();

   while ($line = <$input_file_handle>)
   {
      $cur_line++;

# replace all tabs characters with spaces

      $line =~ s/\t/ /g;

# strip out the CTRL-M's if they exist.  DOS/Windows like to put a CTRL-M at the end of each line

      $line =~ s/\015//;

# Erase any quotation marks, Excel puts them in when you have characters . or , in the cell

      $line =~ s/\"//g;

# Erase all comments

      $line =~ s/\s*#.*$//g;
      chop($line);

      if (index($line, "$_[1]") != -1)
      {
         $no_EOF = 1;
         last;
      }
 
      ${$_[2]} = ${$_[2]}.$line."\n";
   }

   if (($no_EOF == 0) && ($_[0] == 0))
   {
      die "ERROR - Could not find keyword $_[1] in the input file.\n";
   }
 # return comment stripped line
   $line
}

1;
