require parse_main;

########### Table of rules for when phases are allowed to be specified in the Observations

%processor_SPEX_phase_obs = ("u", 0);
%task_SPEX_phase_obs = ("u", 1, "f", 0, "pu", 0);
%entry_SPEX_phase_obs = ("u", 1, "f", 0, "pu", 0, "s", 1, "v", 1, "fb", 0, "pw", 1);
# a info or y info...  No z info, think about how to resolve this problem...
%call_SPEX_phase_obs = ("w", 1);


sub check_SPEX_observations
{
# $_[0]: Reference to an array containing the SPEX observations
# $_[1]: String containing which Block we are in (eg: task, entry, processor... blocks)

# I don't understand why I have to call the import method on parse_main.  According to the book I'm reading
# this is automatically done when you do "use parse_main".

   if ($op_mode eq "SPEX")
   {
      foreach $cur_obs (@{$_[0]}) 
      {
         @key_obs = split(/ +/, $cur_obs);
         $num_key_obs = @key_obs;
         $bare_key = @key_obs[0];

# This expression filters out phase info in the SPEX observational key, and also CORRECT confidence interval possibilities

	 $bare_key =~ s/[0-9]*(\[9[5|9]\])?$//;

         if (($_[1] ne "processor") && ($_[1] ne "task") && ($_[1] ne "call") && ($_[1] ne "entry"))
         {
            die "INTERNAL ERROR - called SPEX observations() with improper parameter $_[1]\n";
         }

         $array_to_reference = $_[1]."_SPEX_phase_obs";

         if ($_[1] eq "call")
         {
            $cur_entity = "entry";
	 }
         else
         {
            $cur_entity = $_[1];
         }

         $error_details = "cur_".$cur_entity;

# If the [xx] confidence interval was not filter out by the $bare_key expression, we will have [xx] present, hence we know the
# confidence interval is incorrect.  If it was correct, all we would be left with is the bare key.  EI: %s[95] -> %s or %s1[95] -> %s

         if (index($bare_key, "[") != -1)
         {  
            die "ERROR at line $cur_line - Invalid confidence interval in SPEX observational key %$bare_key, for $_[1]: $$error_details.\nValid intervals are 95 or 99 percent.\n";
         } 

# Check to see if user specified 2 variables if using the confidence interval, note: num_key_obs = +2 for variables +1 for observation key

         if (($num_key_obs != 3) && (index(@key_obs[0], "[") > -1))
         {
	    die "ERROR at line $cur_line - For %@key_obs[0], you must specify a variable for the observation, and a variable to for the confidence interval.  This is for $_[1]: $$error_details.\n";
	 }

# Check to see if the at least 1 variable, note: num_key_obs = +1 for variables +1 for observation key

         if (($num_key_obs != 2) && (index(@key_obs[0], "[") == -1))
         {
	    die "ERROR at line $cur_line - You need to specify only 1 variable for the %@key_obs[0] observation key, for $_[1]: $$error_details.\n";
         }

         if (not defined $$array_to_reference{$bare_key})
         {
            die "ERROR at line $cur_line - you cannot specify %$bare_key SPEX observation key in the $_[1] block for $cur_entity: $$error_details.\n";
         }

         if ($$array_to_reference{$bare_key} == 0)
         {
            if (@key_obs[0] =~ /[0-9]+(\[.*\])?$/)
            {
	       die "ERROR at line $cur_line - you cannot specify a phase value with observation key %$bare_key in the $_[1] block for $cur_entity: $$error_details.\n";
	    }
         }
      }
   }
}


sub check_model_for_errors
{
   my $vars_match;

   &check_valid_entry_calls(\%call_list_number, 0);
   &check_valid_entry_calls(\%activity_calls, 1);
   &check_calls_to_reference_tasks(\%call_list_number, 0); # check that no entry calls a reference task
   &check_calls_to_reference_tasks(\%activity_calls, 1); # check that no activity calls a reference task
   &check_activity_task_localization;
   &check_activity_connections_in_reference_tasks;
   &check_activity_binding;    # check that at least 1 activity per task is bound to an entry

   if ($disable_cycle_checking == 0)
   {
      &check_for_call_cycles;          # check that entries or activites don't call themselves or have a calling cycle
   }

   if ($op_mode eq "SPEX")
   {
      extract_SPEX_variables(\$SPEX_init_block, \@SPEX_init_block_vars);
      extract_SPEX_variables(\$SPEX_report_block, \@SPEX_report_block_vars);
      extract_SPEX_variables(\$SPEX_convergence_block, \@SPEX_conv_block_vars);

      @SPEX_obs_vars = split(",", $SPEX_obs_var_list);
      $num_obs_vars = @SPEX_obs_vars;

      if ($num_obs_vars == 0)
      {
         die "ERROR - There are no SPEX observation variables present.\n";
      }

      @SPEX_control_vars = split(",", $SPEX_control_var_list);
      $num_control_vars = @SPEX_control_vars;

      if ($num_control_vars == 0)
      {
         die "ERROR - There are no SPEX control variables present.\n";
      }

      if ($debug_msg > 0)
      {
         print "SPEX Init block variables: @SPEX_init_block_vars\n";
         print "SPEX Observation variables: @SPEX_obs_vars\n";
         print "SPEX Control variables: @SPEX_control_vars\n";
         print "SPEX Report block variables: @SPEX_report_block_vars\n";
         print "SPEX Convergence block variables: @SPEX_conv_block_vars\n";
      }

      if ($enable_SPEX_control_dupe_check == 1)
      {
         check_duplicate_SPEX_variables(\@SPEX_control_vars, "control variables");
      }

      check_duplicate_SPEX_variables(\@SPEX_obs_vars, "observation variables");

      $vars_match = "";
      $num_matches = check_SPEX_variables(\@SPEX_control_vars, \@SPEX_init_block_vars, \$vars_match, 0);

      if ($vars_match ne "")
      {
         die "ERROR - Conrtol variables: $vars_match do not appear anywhere in the SPEX initialization block.\n";
      }

      $vars_match = "";
      $num_matches = check_SPEX_variables(\@SPEX_obs_vars, \@SPEX_report_block_vars, \$vars_match, 0);

      if ($vars_match ne "")
      {
         die "ERROR - Observation variables:$vars_match do not appear in the SPEX report block.\n";
      }

# At least 1 control variable and 1 observation or 1 report variable has to be in the convergence block for it to work!

      if ($defined_SPEX_convergence == 1)
      {
         $num_matches = check_SPEX_variables(\@SPEX_conv_block_vars, \@SPEX_control_vars);
         $num_matches1 = check_SPEX_variables(\@SPEX_conv_block_vars, \@SPEX_obs_vars);
         $num_matches2 = check_SPEX_variables(\@SPEX_conv_block_vars, \@SPEX_report_block_vars);

         if (($num_matches == 0) || (($num_matches1 == 0) && ($num_matches2 == 0)))
         {
            die "ERROR - There has to be one SPEX variable from both the observation, and control blocks present in the convergence block\n";
         }
      }

# Check to ensure that no observation and control variables have the same name

      $vars_match = "";
      $num_matches = check_SPEX_variables(\@SPEX_obs_vars, \@SPEX_control_vars, \$vars_match, 1);

      if ($num_matches != 0)
      {
         die "ERROR - The following variables:$vars_match are defined in as control variables and observation variables.\n"; 
      }
   }
}


sub check_valid_entry_calls
{
# $_[0] is a an associative list passed in by reference to check if all the calls in the list are to valid entries.
# $_[1] 0 for checking entries, 1 for checking activities.

# Check to make sure that all calls are made to defined entries
   my $cur_entry = "";
   my $cur_call = "";

   if ($_[1] == 0)
   {
      print "Checking to see if all entry calls made are to defined entries...\n";
   }
   elsif ($_[1] == 1)
   {
      print "Checking to see if all activity calls made are to defined entries...\n";
   }
   else
   {
      die "INTERNAL ERROR - What exactly are you trying to check in check_valid_entry_calls?\n";
   }

   @entries = keys %{$_[0]};

   foreach $cur_entry (@entries)
   {
      @various_calls = keys %{${$_[0]}{$cur_entry}};

      foreach $cur_call (@various_calls)
      {
         @call_parts = split(" ", $_[0]{$cur_entry}{$cur_call});

         if ((@call_parts[0] eq "z") || (@call_parts[0] eq "y"))
         {
            if ((!defined $entry_list_service{@call_parts[2]}) && (!defined $entry_list_activities{@call_parts[2]}))
            {
               if ($_[1] == 0)
               {
                  die "ERROR - entry: $cur_entry is calling entry @call_parts[2] which does not exist\n";
	       }
               else
               {
                  die "ERROR - activity: $cur_entry is calling entry @call_parts[2] which does not exist\n";
               }
	    }
	 }
      }
   }
}

sub check_calls_to_reference_tasks
{
# $_[0] is an associative list passed in by reference that contains a list of all the calls
# $_[1] 0 if checking entires, 1 if checking activities.

   if ($_[1] == 0)
   {
      print "Checking to see if any entry calls a reference task\n";
   }
   elsif ($_[1] == 1)
   {
      print "Checking to see if any activity calls a reference task\n";
   }
   else
   {
      die "INTERNAL ERROR - What are you trying to check with function check_calls_to_reference tasks???\n";
   }

   my $reference_task_exists = 0;

   @tasks = keys %task_list1;
   @calling_entries = keys %{$_[0]};

   foreach $cur_task (@tasks)
   {
      @ref_task_parts = split(" ", $task_list1{$cur_task});

      if (@ref_task_parts[0] eq "r")
      {
         $reference_task_exists = 1;
         $num_ref_entries = @ref_task_parts;

         for ($x=1;$x<$num_ref_entries;$x++)
         {
            foreach $cur_calling_entry (@calling_entries)
            {
#       	print "$x, $call_list{$cur_calling_entry}, @ref_task_parts[$x]\n";

	       if ((defined $entry_list_arriv_rate{@ref_task_parts[$x]}) && ($entry_list_arriv_rate{@ref_task_parts[$x]} != 0))
               {
   	          die "ERROR - Entry \"@ref_task_parts[$x]\" cannot have an open-arrival rate, because it is part of a reference task.\n";
               }

  	       if (defined $_[0]{$cur_calling_entry}{@ref_task_parts[$x]})
               {
                  if ($_[1] == 0)
                  {
                     die "ERROR - Entry: $cur_calling_entry is calling or forwarding to an entry: @ref_task_parts[$x] that belongs to a reference task\n";
		  }
                  else
                  {
                     die "ERROR - Activity: $cur_calling_entry is calling to an entry: @ref_task_parts[$x] that belongs to a reference task\n";
                  }
               }
	    }
         }
      }
   }

   if ($reference_task_exists == 0)
   {
      die "ERROR - No reference tasks have been defined in the model.\n";
   }
}

### Variables below are for check_for_call_cycles.  They are external because check_for_call_cycles is called recursively

$visit_task_list = "";    # string delimited by spaces to keep track what tasks have been visited
$visit_entry_list = "";   # so that we can report exactly the cycle found to the user
%entry_task_lookup = ();
$asynchronous_call_in_cycle_warning = 0;

sub check_for_call_cycles
{
   print "Checking for synchronous call cycles\n";

# It's best to create an entry to task lookup table, this will simplify the code, and speed things up

   @tasks = keys %task_list1;

   foreach $cur_task (@tasks)
   {
      @task_parts = split(" ", $task_list1{$cur_task});

      $num_entries = @task_parts;     # 1 field is the task type, the rest is a list of entries in $task_list1

      for ($x=1;$x<$num_entries;$x++)
      {
         $entry_task_lookup{@task_parts[$x]} = $cur_task;
      }
   }

# We have to add in the activities that are bound to entries into $entry_task_lookup

   @entry_act_list = keys %entry_list_activities;

   foreach $cur_task (keys %task_activity_lookup_list)
   {
      @acts = split(":",  $task_activity_lookup_list{$cur_task});

      foreach $cur_act (@acts)
      {
         foreach $cur_ent (@entry_act_list)
         {
            if ($entry_list_activities{$cur_ent} eq $cur_act)
            {    
                $entry_task_lookup{$cur_ent} = $cur_task;
	    }
         }
      }
   }

# Now we start looking for cycles from every reference task in the model

   foreach $cur_task (@tasks)
   {
      @task_parts = split(" ", $task_list1{$cur_task});

      if (@task_parts[0] eq "r")
      {
         $num_entries = @task_parts;     # 1 field is the task type, the rest is a list of entries in $task_list1

         for ($x=1;$x<$num_entries;$x++)
         {
            $visit_task_list = "";
            $visit_entry_list = "";
            $asynchronous_call_in_cycle_warning = 0;

            &scan_for_cycle(@task_parts[$x]);
         }

   # Now do the same for activities that exist in the reference task
   # code is very similar to what is found in scan_cycle because we can only pass entries into scan_cycles, we can't
   # pass activities.  So for each reference task, need to find all activities which make calls, and start the call
   # trace at the next entry.

	 if (defined $task_activity_lookup_list{$cur_task})
	 {
	    if (index($task_activity_lookup_list{$cur_task}, ":") != -1)
	    {
	       @cur_activity_list = split(":", $task_activity_lookup_list{$cur_task});
	    }
	    else
	    {
	       @cur_activity_list[0] = $task_activity_lookup_list{$cur_task};
	    }

	    foreach my $cur_activity (@cur_activity_list)
	    {
               my @call_parts = keys %{$activity_calls{$cur_activity}};

	       $visit_entry_list = $visit_entry_list." -> (activity) ".$cur_activity;

	       foreach my $cur_call_parts (@call_parts)
	       {
		  my @next_entry = split(" ", $activity_calls{$cur_activity}{$cur_call_parts});

	     # cyclic calls are allowed as long as there is 1 asynchronous call present

		  if (@next_entry[0] eq "z")
		  {
		     $asynchronous_call_in_cycle_warning = 1;
		  }
#		  print "starting an activity jump to another level with @next_entry[2]\n";

		  $visit_task_list = "$cur_task";
		  $visit_entry_list = "(activity) $cur_activity";
		  $asynchronous_call_in_cycle_warning = 0;

		  &scan_for_cycle(@next_entry[2]);
	       }
	    }
	 }
      }
   }
}


sub scan_for_cycle
{
# Note that $_[0] is the current entry this procedure is traversing

    my $visiting_task = $entry_task_lookup{$_[0]};
    my $cur_part="";
    my @cur_activity_list;

    if (index($visit_task_list, $visiting_task) != -1)
    {
       $visit_task_list = $visit_task_list." ".$visiting_task;
       $visit_entry_list = $visit_entry_list." -> ".$_[0];

       if ($asynchronous_call_in_cycle_warning == 0)
       {
          die "ERROR - Found that task: $visiting_task is where synchronous call cycle starts.  Entry cycle in call structure found:\n$visit_entry_list\n";
       }
       else
       {
          print "WARNING - Found that task: $visiting_task is where a call cycle starts, but at least one call is asynchronous.  Entry cycle in call structure found:\n$visit_entry_list\n\n";

          $visit_task_list =~ s/$visiting_task//;
          $visit_entry_list =~ s/ +-> $_[0]//;
          return;
       }
    }

    $visit_task_list = $visit_task_list." ".$visiting_task;

# adds an -> in front of the list, very minor bug.  Only appears when printing out the list.

    $visit_entry_list = $visit_entry_list." -> ".$_[0];

    if (defined $call_list_number{$_[0]})
    {
       my @cur_call_parts = keys %{$call_list_number{$_[0]}};

       foreach $cur_part (@cur_call_parts)
       {
          my @next_entry = split(" ", $call_list_number{$_[0]}{$cur_part});

       # cyclic calls are allowed as long as there is 1 asynchronous call present

          if (@next_entry[0] eq "z")
          {
             $asynchronous_call_in_cycle_warning = 1;
          }
          &scan_for_cycle(@next_entry[2]);
       }
    }

# We can't forget about entries that do forwarding....

    if (defined $call_list_forwarding{$_[0]})
    {
       my @cur_call_parts = keys %{$call_list_forwarding{$_[0]}};

       foreach $cur_part (@cur_call_parts)
       {
          my @next_entry = split(" ", $call_list_forwarding{$_[0]}{$cur_part});

          if (@next_entry[3] != 0)
	  {
             &scan_for_cycle(@next_entry[2]);
	  }
       }
    }

# remove the entry that we just finished with...

#    print "BEFORE we have the visit entry list: $visit_entry_list\n";
    $visit_entry_list =~ s/ +-> $_[0]//;
#    print "AFTER we have the visit entry list: $visit_entry_list\n";

# Now for the current task we are in, go through all the activities, and recursively check for cycles

    if (defined $task_activity_lookup_list{$visiting_task})
    {
       if (index($task_activity_lookup_list{$visiting_task}, ":") != -1)
       {
          @cur_activity_list = split(":", $task_activity_lookup_list{$visiting_task});
       }
       else
       {
          @cur_activity_list[0] = $task_activity_lookup_list{$visiting_task};
       }

       foreach my $cur_activity (@cur_activity_list)
       {
          my @call_parts = keys %{$activity_calls{$cur_activity}};

          $visit_entry_list = $visit_entry_list." -> (activity) ".$cur_activity;

	  foreach my $cur_call_parts (@call_parts)
	  {
	     my @next_entry = split(" ", $activity_calls{$cur_activity}{$cur_call_parts});

	  # cyclic calls are allowed as long as there is 1 asynchronous call present

	     if (@next_entry[0] eq "z")
	     {
		$asynchronous_call_in_cycle_warning = 1;
	     }
#	     print "jumping to another level with @next_entry[2]\n";

	     &scan_for_cycle(@next_entry[2]);
             $visit_entry_list =~ s/ +-> \(activity\) $cur_activity//;
	  }
       }
   }

# remove ourselves from the visited list, otherwise this algorithm will detect a cycle
# incorrectly if an entry *eventually* calls the same task twice.

    $visit_task_list =~ s/$visiting_task//;
}


sub check_activity_task_localization
{
   my @tasks_with_connections = keys %activity_connections;

   print "Checking to ensure that all activities are localized to only 1 task...\n";

   foreach my $cur_task (@tasks_with_connections)
   {
# First we should check that where ever there are activity connections, activities exist in the task!

      if (!defined $task_activity_lookup_list{$cur_task})
      {
         die "ERROR - In task $cur_task, an activity connection statement is present, but there are no activities in the task!\n";
      }

      $cur_connection = $activity_connections{$cur_task};

   # To make life easier for the regexps, we'll insert 1 space into the string we are going to filter.

      $cur_connection = " ".$cur_connection;

   # Need to extract the words out of the connection statement, so we need to filter out various characters

      $cur_connection =~ s/\s*[\&\+\*\(\)\,\;\:\[\]\n]/ /g;
      $cur_connection =~ s/\s*->/ /g;
      $cur_connection =~ s/\s+[0-9]+\s+/ /g;
      $cur_connection =~ s/\s+[0-9]*[\.]?[0-9]*/ /g;
      $cur_connection =~ s/\s+[Ee][\-]?[0-9]+/ /g;
      
   # get rid of space in the front of the string if there is any

      $cur_connection =~ s/^\s+//;

      @entities = split(" ", $cur_connection);

      foreach $cur_entity (@entities)
      {
   # Check that all the activities in the activity connection belong in the same task

         if (index($task_activity_lookup_list{$cur_task}, $cur_entity.":") == -1)
         {
   # Check if the activity is defined anywhere in the model...

            if (defined $activity_list_service{$cur_entity})
            {
               die "ERROR - In the activity connection statement for task \"$cur_task\", activity \"$cur_entity\" is defined in another task.\n";
            }
            
            if (index($task_list1{$cur_task}, $cur_entity) == -1)
            {
	       if (defined $entry_list_service{$cur_entity})
               {
       	          die "ERROR - In the activity connection statement for task \"$cur_task\", entry \"$cur_entity\" is defined in another task.\n";
               }
               else
               {
	          die "ERROR - In the activity connection statement for task \"$cur_task\", a reference to an entry/activity \"$cur_entity\" does not even exist in the model.\n";
	       }
            }
	 }
      }
   }
}


sub check_activity_connections_in_reference_tasks
{
   print "Checking reference tasks if they have any activity connections...\n";

   foreach my $cur_task (keys %task_list1)
   {
      @ref_task_parts = split(" ", $task_list1{$cur_task});

      if (@ref_task_parts[0] eq "r")
      {
         if (defined $activity_connections{$cur_task})
         {
            print "WARNING - Task \"$cur_task\" is a reference task that has activity connections.\n";
         }
      }
   }
}

sub check_activity_binding
{
   my @tasks = keys %task_list1;

   foreach my $cur_task (@tasks)
   {   
      @activities = split(":", $task_activity_lookup_list{$cur_task});

      if (@activities[0] ne "")
      {
         $binding_exists = 0;

         foreach my $cur_act (@activities)
         {
	    @entry_activity_list = keys %entry_list_activities;

            foreach my $cur_entry (@entry_activity_list)
            {
   	       if ($cur_act eq $entry_list_activities{$cur_entry})
               {
                  $binding_exists = 1;
                  last;
	       }
            }
            if ($binding_exists == 1)
            {
               last;
            }
         }
      
         if ($binding_exists == 0)
         {
            die "ERROR - Task $cur_task has activities, but non of them are bound to any entries.\n";
         }
      }
   }
}

sub check_valid_identifier_name
{
   my $value = $_[0];

   if (!($value =~ /^_*[a-zA-Z]+[a-zA-Z_0-9\-\/]*$/))
   {
      die "ERROR at line $cur_line: \"$_[0]\" is not a valid identifier\n";
      exit 1;
   }
}

sub check_valid_real
{
   my $value = $_[0];

  
  if (!($value =~ /^\+?\.?[0-9]+\.?[0-9]*[\s]*([eE]?[\s]*[\+\-]?[0-9]+)?$/))
  {
     if ($op_mode eq "SPEX")
     {
        if (index($_[0], "\$") != 0)
        {
           die "ERROR at line $cur_line: \"$_[0]\" is not a SPEX variable or a positive real number.\n";
        }
     }
     else
     {
        die "ERROR at line $cur_line: \"$_[0]\" is not a positive real number\n";
     }
  }
}

sub check_valid_integer
{
   my $value = $_[0];

  if (!($value =~ /^[0-9]+$/))
  {
     if ($op_mode eq "SPEX")
     {
        if (index($_[0], "\$") != 0)
        {
           die "ERROR at line $cur_line: \"$_[0]\" is not a SPEX variable or a positive integer number.\n";
        }
     }
     else
     {
        die "ERROR at line $cur_line: \"$_[0]\" is not a positive integer number.\n";
     }
  }
}

sub check_SPEX_variables
{
# $_[0] variable reference to an array of SPEX variables.  All these variables will be searched to see if
#       they exist in $_[1].
# $_[1] variable reference to an array of another variable list to compare with $_[0]
# $_[2] a reference string which contains the variables that matched, or didn't match
# $_[3] 1: return list in $_[2] that matched, 0: return list in $_[1] that didn't match

   my $var_list0 = $_[0];
   my $var_list1 = $_[1];
   my $num_params = @_;
   my $found_match;

   if (($num_params !=2) && ($num_params != 4))
   {  
      die "INTERNAL ERROR - Did not pass 2 or 4 parameters into function check_SPEX_variables.\n";
   }
   
   $search_match = 0;

   foreach $cur_var0 (@{$var_list0})
   {
       $found_match = 0;

       foreach $cur_var1 (@{$var_list1})
       {
          if ($cur_var0 eq $cur_var1)
          {
	     $search_match++;
             $found_match = 1;

             if (($num_params == 4) && ($_[3] == 1))
             {
                ${@_[2]} = ${@_[2]}." ".$cur_var0;
	     }
          }
       }
       if (($found_match == 0) && ($num_params == 4) && ($_[3] == 0))
       {
          ${@_[2]} = ${@_[2]}." ".$cur_var0;
       }
   }
   $search_match;
}


sub check_duplicate_SPEX_variables
{
# $_[0]: Reference to an array of variables
# $_[1]: A string which contains a description about the array of variables (for making informative error messages)

   my @var_list = @{$_[0]};
   my $cur_var, $check_var;
   
   while ($#var_list != -1)
   {
      $check_var = shift @var_list;
      
      foreach $cur_var (@var_list)
      {
         if ($cur_var eq $check_var)
         {
            die "ERROR - there is a duplicate variable: $cur_var in the SPEX $_[1].\n";
         }
      }
   }
}

# This has to be here for some reason, so it returns a TRUE.
1;
