require parse_main;

$output_file_handle;

$DEFAULT_CONVERGENCE_LIMIT = .00001;
$DEFAULT_ITERATION_LIMIT = 100;
$DEFAULT_PRINT_INTERVAL = 1;
$DEFAULT_UNDERRELAXATION_COEFF = 0.9;
$DEFAULT_COV = 1;     # exponential
$DEFAULT_PHASE = 0;   # stochastic
$DEFAULT_FANIN = 0;
$DEFAULT_FANOUT = 0;
$DEFAULT_THINKTIME = 0;
$DEFAULT_MAXSERVICE = 0;


sub generate_LQN_file
{
# $_[0]: The name of the file that which will contain the output of the parser.

   if (open (OUTPUT_FILE, ">".$_[0].".lqn") != 1)
   {
      die "ERROR - cannot open file \"$_[0]\" for writing.\n";
   }

   $output_file_handle = *OUTPUT_FILE;


   print "WARNING - Only first phase value of Fan In and Fan Out will be read in.  Currently LQNS cannot handle a vector of values for those two parameters.\n";

   &generate_general_info_block;
   &generate_processor_block;
   &generate_task_block;
   &generate_entry_block;
   &generate_activity_block;
   close($output_file_handle);
}

sub generate_SPEX_file
{
   if (open (OUT_LQN, ">".$_[0].".xlqn") != 1)
   {
      die "ERROR - cannot open file \"$output_file\" for writing.\n";
   }


   print "WARNING - Only first phase value of Fan In and Fan Out will be read in.  Currently LQNS cannot handle a vector of values for those two parameters.\n";

   $output_file_handle = *OUT_LQN;

   &generate_SPEX_init_block;
   &generate_processor_block;
   &generate_task_block;
   &generate_entry_block;
#   &generate_activity_block;
   &generate_SPEX_report_block;

   if ($defined_SPEX_convergence == 1)
   {
      &generate_SPEX_convergence_block;
   }
   close($output_file_handle);
}

sub generate_SPEX_init_block
{
#   print $output_file_handle "$model_comments\n";

   if ($convergence_limit != $DEFAULT_CONVERGENCE_LIMIT)
   {
      print $output_file_handle "\$convergence_limit = $convergence_limit;\n";
   }

   if ($iteration_limit != $DEFAULT_ITERATION_LIMIT)
   {
      print $output_file_handle "\$iteration_limit = $iteration_limit\n";
   }

   if ($print_interval != $DEFAULT_PRINT_INTERVAL)
   {
      print $output_file_handle "\$print_interval = $print_interval\n";
   }
   
   if ($underrelaxation_coeff != $DEFAULT_UNDERRELAXATION_COEFF)
   {
      print $output_file_handle "\$coefficient_of_variation = $underrelaxation_coeff\n\n";
   }

   @pragmas = split("\n", $model_pragmas);

   foreach $cur_pragma (@pragmas)
   {
      $cur_pragma =~ s/\s*//g;
      if ($cur_pragma ne "")
      {
         print $output_file_handle "pragma $cur_pragma\n";
      }
   }
   print $output_file_handle "\n";

   print $output_file_handle "$SPEX_init_block\n";
}

sub generate_SPEX_report_block
{
   print $output_file_handle "R 0\n";
   print $output_file_handle "$SPEX_report_block";
   print $output_file_handle "-1\n\n";
}

sub generate_SPEX_convergence_block
{
   print $output_file_handle "C 0\n";
   print $output_file_handle "$SPEX_convergence_block";
   print $output_file_handle "-1\n\n";
}

sub generate_general_info_block
{
   print $output_file_handle "G\n";
   print $output_file_handle "\"$model_comments\"\n";
   print $output_file_handle "$convergence_limit\n";
   print $output_file_handle "$iteration_limit\n";
   print $output_file_handle "$print_interval\n";
   print $output_file_handle "$underrelaxation_coeff\n";
   print $output_file_handle "-1\n\n";

   @pragmas = split("\n", $model_pragmas);

   foreach $cur_pragma (@pragmas)
   {
      $cur_pragma =~ s/\s*//g;
      if ($cur_pragma ne "")
      {
         print $output_file_handle "\#pragma $cur_pragma\n";
      }
   }
   print $output_file_handle "\n";
}

sub generate_processor_block
{
   $num_processors = keys %processor_list;
   print $output_file_handle "P $num_processors\n";

   foreach $processor (keys %processor_list)
   {
      if ($op_mode eq "SPEX")
      { 
         @SPEX_info = split(";", $processor_list{$processor});
         @temp_info = split(" ", @SPEX_info[0]);
      }
      else
      {      
         @temp_info = split(" ", $processor_list{$processor});
      }
     
      print $output_file_handle "p $processor @temp_info[0]";

   # Check to see if Quantum is a non-zero value

      if ((@temp_info[2] != 0) || (index(@temp_info[2], "\$") == 0))
      {
         print $output_file_handle " @temp_info[2]";
      }

# if there is only 1 copy, no need to add in the "m <value>" part in the LQN file

   # Below we have to qualify with checking "\$", because a string when checked against a numeric value equals 0
      if (((@temp_info[1] eq "i") || (@temp_info[1] == 0)) && (index(@temp_info[1], "\$") != 0))
      {
         print $output_file_handle " i";
      }
      elsif ((@temp_info[1] > 1) || (index(@temp_info[1], "\$") == 0))
      {
         print $output_file_handle " m @temp_info[1]";
      }

   # Check to see if Replicas is any value other than 0

      if ((@temp_info[3] != 0) || (index(@temp_info[3], "\$") == 0))
      {
         print $output_file_handle " r @temp_info[3]";
      }

   # Check to see if Proc_Rate is any value other than 1

      if ((@temp_info[4] != 1) || (index(@temp_info[4], "\$") == 0))
      {
         print $output_file_handle " R @temp_info[4]";
      }

      if ($op_mode eq "SPEX")
      {
         if ($SPEX_info[1] ne "")
         {
            print $output_file_handle "    $SPEX_info[1]";
         }
      }
      print $output_file_handle "\n";
   }
   print $output_file_handle "-1\n\n";
}

sub generate_task_block
{
   $num_tasks = keys %task_list1;

   print $output_file_handle "T $num_tasks\n";

   foreach $task (keys %task_list1)
   {
      if ($op_mode eq "SPEX")
      { 
         @SPEX_info = split(";", $task_list2{$task});
         @temp_info = split(" ", @SPEX_info[0]);
      }
      else
      {      
         @temp_info = split(" ", $task_list2{$task});
      }

      print $output_file_handle "t $task $task_list1{$task}-1 @temp_info[0]";

# Check to see if task priority is 0, if not then output the priority into the LQN file

     if ((@temp_info[1] != 0) || (index(@temp_info[1], "\$") == 0))
     {
        print $output_file_handle " $temp_info[1]";
     }

# Check to see if the think time is 0, if not then output the think time to the LQN file

     if ((@temp_info[2] != 0) || (index(@temp_info[2], "\$") == 0))
     {
        print $output_file_handle " z $temp_info[2]";
     }

# if there is only 1 copy, no need to add in the "m <value>" part in the LQN file

     if (@temp_info[3] == 1)
     {
        print $output_file_handle "";
     }
     elsif (((@temp_info[3] eq "i") || (@temp_info[3] == 0)) && (index(@temp_info[3], "\$") != 0)) # deal with infinite multiplicity
     {
        print $output_file_handle " i";
     }
     else  # we have finite mulitplicity
     {
        print $output_file_handle " m @temp_info[3]";
     }

# Check to see if the replica field is 0, if not then output to the LQN file

     if ((@temp_info[4] != 0) || (index(@temp_info[4], "\$") == 0))
     {
        print $output_file_handle " r $temp_info[4]";
     }

     if ($op_mode eq "SPEX")
     { 
         if ($SPEX_info[1] ne "")
         {
            print $output_file_handle "    $SPEX_info[1]";
         }
     }

     print $output_file_handle "\n";
   }
   print $output_file_handle "-1\n\n";
}

sub generate_entry_block
{
   $num_entries = keys %entry_list_service;
   $num_entries = $num_entries + keys %entry_list_activities;

   print $output_file_handle "E $num_entries\n";

   foreach $entry (keys %entry_list_service)
   {      
# specify HOL entry priority only if the priority number is not 0 (zero), or it's a SPEX variable

      if (($entry_list_priority{$entry} != 0) || (index($entry_list_priority{$entry}, "\$")) == 0)
      {
         print $output_file_handle "p $entry $entry_list_priority{$entry}\n";
      }

# specify an open arrival rate only if it's greater than 0, or it's a SPEX variable

      if (($entry_list_arriv_rate{$entry} != 0) || (index($entry_list_arriv_rate{$entry}, "\$") == 0))
      {
         print $output_file_handle "a $entry $entry_list_arriv_rate{$entry}\n";
      }

      @SPEX_info = split(";", $entry_list_service{$entry});
      @entry_service_halves = split("!", @SPEX_info[0]);

      print $output_file_handle "@entry_service_halves[0]$entry@entry_service_halves[1] -1";

      if ($op_mode eq "SPEX")
      { 
         if ($SPEX_info[1] ne "")
         {
            print $output_file_handle "    $SPEX_info[1]";
         }
      }

      print $output_file_handle "\n";

# output only if they specify a COV, default COV is 1 (exponential), 0 for deterministic

      if ((($op_mode eq "SPEX") && (index($entry_list_COV{$entry}, "\$") != -1)) || (&all_default_values($entry_list_COV{$entry}) != 1))
      {
	 @entry_COV_halves = split("!", $entry_list_COV{$entry});
         print $output_file_handle "@entry_COV_halves[0]$entry@entry_COV_halves[1] -1\n";
      }

# output only if user specifies a non default phasetype, or puts in a SPEX variable, default is STOCHASTIC meaning 0, DETERMINISTIC is 1

      if ((($op_mode eq "SPEX") && (index($entry_list_phasetype{$entry}, "\$") != -1)) || (&all_default_values($entry_list_phasetype{$entry}) != 1))
      {
          @entry_phasetype_halves = split("!", $entry_list_phasetype{$entry});
          print $output_file_handle "@entry_phasetype_halves[0]$entry@entry_phasetype_halves[1] -1\n";
      }

# output only if user specifies a non-zero Think_Time, or puts in a SPEX variable.

      if ((($op_mode eq "SPEX") && (index($entry_list_thinktime{$entry}, "\$") != -1)) || (&all_default_values($entry_list_thinktime{$entry}) != 1))
      {
          @entry_thinktime_halves = split("!", $entry_list_thinktime{$entry});
          print $output_file_handle "@entry_thinktime_halves[0]$entry@entry_thinktime_halves[1] -1\n";
      }

# output only if user specifies a non-zero Max_Service, or puts in a SPEX variable.

      if ((($op_mode eq "SPEX") && (index($entry_list_maxservice{$entry}, "\$") != -1)) || (&all_default_values($entry_list_maxservice{$entry}) != 1))
      {
          @entry_maxservice_halves = split("!", $entry_list_maxservice{$entry});
          print $output_file_handle "@entry_maxservice_halves[0]$entry@entry_maxservice_halves[1] -1\n";
      }

      &generate_entry_calls($entry);
   }

   &generate_entry_activity_bindings;

   print $output_file_handle "-1\n\n";
}


sub generate_entry_calls
{
# passing the name of the current entry into this function which is $_[0]

   my $cur_entry = $_[0];

   @call_parts = keys %{$call_list_number{$cur_entry}};

   foreach $cur_call (@call_parts)
   {

# NOTE: here we have to split the LQN entry information line in half to insert the entry name.

      @SPEX_info = split(";", $call_list_number{$cur_entry}{$cur_call});
      @call_halves = split("!", @SPEX_info[0]);

      print $output_file_handle "@call_halves[0]$cur_entry@call_halves[1] -1";

      if ($op_mode eq "SPEX")
      { 
         if ($SPEX_info[1] ne "")
         {
            print $output_file_handle "    $SPEX_info[1]";
         }
      }

      print $output_file_handle "\n";
   }

   @call_parts = keys %{$call_list_fanin{$cur_entry}};

   foreach $cur_call (@call_parts)
   {
      $cur_call_part = $call_list_fanin{$cur_entry}{$cur_call};

# output only if user specifies a non-zero Fan_In, or puts in a SPEX variable.

      if ((($op_mode eq "SPEX") && (index($cur_call_part, "\$") != -1)) || (&all_default_values($cur_call_part) != 1))
      {
          @call_fanin_halves = split("!", $cur_call_part);
          print $output_file_handle "@call_fanin_halves[0]$entry@call_fanin_halves[1] -1\n";
      }
   }

   @call_parts = keys %{$call_list_fanout{$cur_entry}};

   foreach $cur_call (@call_parts)
   {
      $cur_call_part = $call_list_fanout{$cur_entry}{$cur_call};

# output only if user specifies a non-zero Fan_Out, or puts in a SPEX variable.

      if ((($op_mode eq "SPEX") && (index($cur_call_part, "\$") != -1)) || (&all_default_values($cur_call_part) != 1))
      {
          @call_fanout_halves = split("!", $cur_call_part);
          print $output_file_handle "@call_fanout_halves[0]$entry@call_fanout_halves[1] -1\n";
      }
   }

   @call_parts = keys %{$call_list_forwarding{$cur_entry}};

   foreach $cur_call (@call_parts)
   {
      $cur_call_part = $call_list_forwarding{$cur_entry}{$cur_call};

# output only if user specifies a non-zero Fan_Out, or puts in a SPEX variable.

      if ((($op_mode eq "SPEX") && (index($cur_call_part, "\$") != -1)))
      {
          @call_forward_halves = split("!", $cur_call_part);
          print $output_file_handle "@call_forward_halves[0]$entry@call_forward_halves[1] -1\n";
      }
   }
}


sub generate_entry_activity_bindings
{
   foreach $act_entry (keys %entry_list_activities)
   {
      print $output_file_handle "A $act_entry $entry_list_activities{$act_entry}\n";
   }
}


sub generate_activity_block
{
#   $num_activities = keys %activity_list_service;
   @tasks_with_activities = keys %task_activity_lookup_list;

   foreach my $cur_task (@tasks_with_activities)
   {
      my @activity_list = split(":", $task_activity_lookup_list{$cur_task});

      print $output_file_handle "A $cur_task\n";

      foreach $activity (@activity_list)
      {      
   # output service time for the activity.

	 if (($activity_list_service{$activity} != 0) || (index($activity_list_service{$activity}, "\$") == 0))
	 {
	    print $output_file_handle "s $activity $activity_list_service{$activity}\n";
	 }

   # specify Think_Time for activity only if it's non-zero, or it's a SPEX variable

	 if (($activity_list_thinktime{$activity} != 0) || (index($activity_list_thinktime{$activity}, "\$")) == 0)
	 {
	    print $output_file_handle "Z $activity $activity_list_thinktime{$activity}\n";
	 }

   # output only if they specify a COV, default COV is 1 (exponential), 0 for deterministic

	 if (($activity_list_COV{$activity} != 1) || (index($activity_list_COV{$activity}, "\$") == 0))
	 {
	    print $output_file_handle "c $activity $activity_list_COV{$activity}\n"
	 }

   # output only if user specifies a non default phasetype, or puts in a SPEX variable, default is STOCHASTIC meaning 0, DETERMINISTIC is 1

	 if (($activity_list_phasetype{$activity} != 0) || (index($activity_list_phasetype{$activity}, "\$") == 0))
	 {
	     print $output_file_handle "f $activity $activity_list_phasetype{$activity}\n";
	 }

	 &generate_activity_calls($activity);
      }

      if (defined $activity_connections{$cur_task})
      {
         print $output_file_handle ":\n";
	 print $output_file_handle "$activity_connections{$cur_task}\n";
      }

      print $output_file_handle "-1\n\n";
   }
}


sub generate_activity_calls
{
# passing the name of the current activity into this function which is $_[0]

   my $cur_activity = $_[0];

   @call_parts = keys %{$activity_calls{$cur_activity}};

   foreach $cur_call_part (@call_parts)
   {
      @call_halves = split("!", $activity_calls{$cur_activity}{$cur_call_part});

      print $output_file_handle "@call_halves[0]$cur_activity@call_halves[1]\n";
   }
}


sub all_default_values
{
# This routine checks a given LQN formatted line to see if there are any default values present which may not need printing.

   my @parts = split(" ", $_[0]);
   my $num_parts = @parts;
   my $default_result = 0;
   my $default_value;
   my $start_place;

   if ((@parts[0] ne "z") && (@parts[0] ne "y"))
   {
      if (@parts[0] eq "f")   # checking to see if all phasetypes are stochastic (0)
      {
         $default_value = $DEFAULT_PHASE;
         $start_place = 2;
      }
      elsif (@parts[0] eq "c") # checking to see if all COV are exponential (1)
      {
         $default_value = $DEFAULT_COV;   
         $start_place = 2;
      }
      elsif (@parts[0] eq "i")  # checking to see if all Fan_In is 0
      {
         $default_value = $DEFAULT_FANIN;
         $start_place = 3;
      }
      elsif (@parts[0] eq "o")  # checking to see if all Fan_Out is 0
      {
         $default_value = $DEFAULT_FANOUT;
         $start_place = 3;
      }
      elsif (@parts[0] eq "Z") # checking to see if Think_Time is 0
      {
         $default_value = $DEFAULT_THINKTIME;
         $start_place = 2;
      }
      elsif (@parts[0] eq "M") # checking to see if Max_Service is 0
      {
         $default_value = $DEFAULT_MAXSERVICE;
         $start_place = 2;
      }
     else
     {
        die "INTERNAL ERROR - spreadsheet parser has reached an unknown state in function all_default_values, looking for default values for parameter: \"@parts[0]\".\n";
     }

# we start looking through the fields to see if certain values are all the same

      for ($x=$start_place;$x<$num_parts;$x++)
      {
         if (@parts[$x] != $default_value)
         {
            last;
         }
      }

      if ($x == $num_parts)
      {
         $default_result = 1;
      }
   }
   $default_result;
}

# We need to return a TRUE (1), otherwise Perl complains.
1;
